#wx_trace_flag_box.py   18Apr2024  crs Author
""" Select Trace Flag data display
Facilitates setting and display of select flag value
design
flag -text
value - text  with type: boolean, int
    types:              display/input
            boolean     check box
            int         entry
            string      entry
display
    [ ] flag name
        
"""
import re
import wx
from select_trace import SlTrace

class TraceFlagBox(wx.Panel):
    def __init__(self,
                 parent : wx.Panel,
                 flag : str,
                 typ : type = None,
                 value = None):
        """ Setup entry/display of a flag
        :flag: name of flag
        :typ: type of value default: Determined by value if present
                                    else: bool
        :value: current value default:
        """
        super().__init__(parent)
        self.flag = flag
        if typ is None:
            if value is not None:
                typ = type(value)
            else:
                typ = bool                
        self.type = typ
        if value is None:
            if typ == bool:
                value = False
            elif typ == int:
                value = 0
            elif typ == float:
                value = 0.
            else:
                value = ""
        self.value = value
        
        tfb_sizer = wx.BoxSizer(wx.HORIZONTAL)
        if typ == bool:
            val_ctl = wx.CheckBox(self, label=flag)   # Treat all flags the same -
            val_ctl.SetValue(value)
            self.Bind(wx.EVT_CHECKBOX, self.on_checkbox)                                        #   sizer
        elif typ == int:
            val_ctl = wx.TextCtrl(self, value=str(value))
        elif typ == float:
            val_ctl = wx.TextCtrl(self, value=str(value))
        elif typ == str:
            val_ctl =  wx.TextCtrl(self, value=value)
        else:
            raise Exception(f"Unrecognized type {typ}")
        self.val_ctl = val_ctl
        tfb_sizer.Add(val_ctl, 0, wx.RIGHT)     # checkbox has own label
        if typ != bool:
            tfb_sizer.Add(wx.StaticText(self, label=flag), 0, wx.RIGHT)
        self.SetSizer(tfb_sizer)

    def on_checkbox(self, event):
        """Process any checkbox click
        :event: event
        """
        cb = event.GetEventObject()
        flag = cb.GetLabel()
        value = cb.GetValue()
        self.set(flag, value)
        
    def get(self):
        """ Get current value
        """
        return self.value

    def set(self, flag, val):
        """ Set value
        :flag: flag string
        :val: new value
        """
        SlTrace.setTraceFlag(flag, val)
        if isinstance(self.val_ctl, wx.CheckBox):    
            self.val_ctl.SetValue(val)
        elif isinstance(self.val_ctl, wx.TextCtrl):
            self.val_ctl = wx.TextCtrl(self, value=str(val))
            
            
    def str_to_type(self, st):
        """ Determine flag type value string
        :st: value string
        :returns: variable type default: bool
        """
        
        if st is None:
            return bool
        if type(st) == str:
            stl = st.lower()
            if stl in ["t", "true", "1", "f", "false", "0"]:
                return bool
            
            if re.match(r"^\d+", st):
                return int
        
            if re.match(r"^((\d+\.\d*)|(\d*\.\d+)$", st):
                return float
        
        return str       # else Assume a string
        
        
            
                
if __name__ == "__main__":
    SlTrace.lg("Testing")
    
    app = wx.App()
    frame = wx.Frame(None)
    panel = wx.Panel(frame)
    tfl_sizer = wx.BoxSizer(wx.VERTICAL)
    panel.SetSizer(tfl_sizer)

    flags = SlTrace.getAllTraceFlags()
    for flag in flags:
        value = SlTrace.getLevel(flag)
        tfb = TraceFlagBox(panel, flag, value=value)
        tfl_sizer.Add(tfb, proportion=1)
    frame.Show()
    
    app.MainLoop()
    


#tk_canvas_grid.py  29Dec2023   crs  Pull from canvas_grid.py 
#                   14Feb2023  crs  Author
"""
Basis for nestable grid layout within the standard tkinter Canvas Widget.
Facilitate creating coarse picture from tkinter Canvas contents

"""
import tkinter as tk
import sys
import os
import copy

from select_trace import SlTrace
from braille_error import BrailleError
from magnify_info import MagnifySelect, MagnifyInfo, MagnifyDisplayRegion

"""
We now think explicit .base.fn_name is better
for indirect call of tk.Canvas calls:
    find_overlapping, gettags, itemconfigure, type
"""

class TkCanvasGrid(tk.Canvas):
        
    def __init__(self, master, base=None,
                 g_xmin=None, g_xmax=None, g_ymin=None, g_ymax=None,
                 g_nrows=25, g_ncols=40,
                 **kwargs):
        """ Set up canvas object with grid
        :base: tk.Canvas, if present, from which we get
                canvas item contents
                default: create tk.Canvas(master)
        :g_xmin,g_max: if None g_xmin = -width/2, g_x_max = +width/2
        :g_xmin: Grid minimum canvas coordinate value default: left edge
        :g_xmax: Grid maximum canvas coordinate value default: right edge
        :g_ymin: Grid minimum canvas coordinate value default: top edge
        :g_ymax: Grid maximum canvas coordinate value default: bottom edge
        :g_nrows: Number of rows default: 25
        :g_ncols: Number of columns default: 40
        """
        have_base = True    # True iff we have a base
        if base is None:
            have_base = False
            super().__init__(master=master, **kwargs)
            self.pack(expand=True, fill=tk.BOTH)
            self.master.update()
            base = self
        else:
            super().__init__(master=master, **kwargs)
        self.base = base
        self.item_samples = {}      # For incremental presentation  via show_item
        self.n_cells_created = 0    # Number of cells in recent window
        if g_xmin is None and g_xmax is None:
            g_xmin = -base.winfo_width()/2
            g_xmax = -g_xmin 
        if g_xmin is None:
            g_xmin = 0
        self.g_xmin = g_xmin
        if g_xmax is None:
            g_xmax = g_xmin + base.winfo_width()
        self.g_xmax = g_xmax
        self.g_width = g_xmax-g_xmin
        
        if g_ymin is None and g_ymax is None:
            g_ymin = -base.winfo_height()/2
            g_ymax = -g_ymin 
        if g_ymin is None:
            g_ymin = 0
        self.g_ymin = g_ymin
        if g_ymax is None:
            g_ymax = g_ymin + base.winfo_height()
        self.g_ymax = g_ymax
        self.g_height = abs(g_ymax-g_ymin) # No questions
        self.g_nrows = g_nrows
        self.g_ncols = g_ncols
        self.grid_tag = None        # Most recent grid paint tag
        self.set_grid_lims()
        
    def get_grid_lims(self, xmin=None, xmax=None, ymin=None, ymax=None,
                      ncols=None, nrows=None):
        """ create grid boundary values bottom through top, given limits
         so:
             grid_xs[0] == left edge
             grid_xs[grid_width] == right edge
             grid_ys[0] == top edge
             grid_ys[grid_height] == bottom edge
             cell(i,j): grid_xs[i], grid_xs[i+1], grid_ys[j], grid_ys[j+1]
            All coordinates are in canvas values: y==0 at top, increasing down
        :xmin: minimum x default: self.g_xmin
        :xmax: maximim x default: self.g_xmax
        :ymin: minimum y default: self.g_ymin
        :ymax: maximim y defailt: self.g_ymax
        :ncols: grid horizontal columns default: self.g_ncols
        :nrows: grid vertical rows default: self.g_nrows
        :returns: (grid_xs,grid_ys) - list of x limits, list of y limits
        """
        if xmin is None:
            xmin = self.g_xmin 
        if xmax is None:
            xmax = self.g_xmax 
        if ymin is None:
            ymin = self.g_ymin 
        if ymax is None:
            ymax = self.g_ymax 
        if ncols is None:
            ncols = self.g_ncols 
        if nrows is None:
            nrows = self.g_nrows 
            
        g_width = abs(xmax-xmin)
        g_height = abs(ymax-ymin) 
        grid_xs = []
        grid_ys = []

        for i in range(ncols+1):
            x = xmin + i*g_width/ncols
            grid_xs.append(x)
        for i in range(nrows+1):
            y = ymin + i*g_height/nrows
            grid_ys.append(y)
        return grid_xs,grid_ys
        
    def set_grid_lims(self):
        """ create grid boundary values bottom through top
         so:
             grid_xs[0] == left edge
             grid_xs[grid_width] == right edge
             grid_ys[0] == top edge
             grid_ys[grid_height] == bottom edge
             cell(i,j): grid_xs[i], grid_xs[i+1], grid_ys[j], grid_ys[j+1]
        """
        self.grid_xs, self.grid_ys = self.get_grid_lims()

    def item_to_color(self, item_ids):
        """ Get color string given item id
            Ignore items with fill tuple ending with ''
        :item_ids: list of canvas ids for item
        :returns: color str or None if no item with valid color
        """
        for top_id in reversed(item_ids):
            color_tuple = self.base.itemconfigure(top_id, "fill")
            color = color_tuple[-1]
            if color != '':
                return color       # Take first that has a color  not ''
        return None
        
    def display_cell(self, ixy_item, mark_type=None):
        """ Display grid square based on graphics
        :ixy: (ix,iy) tuple for location
        :mark_type: type one of "braille" - simulated braille
                                "square" - square of cell
                    default: "square"
        """
        (ix,iy), ids = ixy_item
        
        if mark_type is None:
            mark_type = "square"
        color = self.item_to_color(item_ids=ids)
        if color is not None:
            w_left_x,w_upper_y, w_right_x,w_lower_y = self.get_grid_ullr(ix=ix, iy=iy)
            self.create_rectangle(w_left_x, w_upper_y, w_right_x, w_lower_y,
                                  outline=color, width=3)

    def display_cells(self):
        items = self.get_canvas_items()
        for item in items:
            self.display_cell(item)
                    
    def paint_grid(self, color="gray", grid_tag="grid_tag", width=2):
        """ paint grid lines for diagnostic purposes
        :color: grid line color default:gray
        :grid_tag: tag for grid canvas items
        :width: line width default: 2
        """
        self.grid_tag = grid_tag
        width_edge = 7         # So it shows up
        width_now = width
        for i in range(len(self.grid_xs)):          # vertical lines
            x1,x2 = self.grid_xs[i],self.grid_xs[i]
            y1,y2 = self.grid_ys[0],self.grid_ys[len(self.grid_ys)-1]
            width_now = width_edge if i == 0 or i == len(self.grid_xs)-1 else width
            self.create_line(x1,y1,x2,y2, fill=color, tag=grid_tag, width=width_now)
            self.update()
            pass
        for i in range(len(self.grid_ys)):        # horizontal lines
            y1,y2 = self.grid_ys[i],self.grid_ys[i]
            x1,x2 = self.grid_xs[0],self.grid_xs[len(self.grid_xs)-1]
            width_now = width_edge if i == 0 or i == len(self.grid_ys)-1 else width
            self.create_line(x1,y1,x2,y2, fill=color, tag=grid_tag, width=width_now)
            
        self.update()

    def erase_grid_paint(self, grid_tag=None):
        """ Erase painted grid
        :grid_tag: grid_tag used default: self.grid_tag - last grid painted
        """
        if grid_tag is None:
            grid_tag = self.grid_tag
        if grid_tag is not None:
            self.delete(grid_tag)
            self.grid_tag = None
            self.update()

                       
    def get_grid_ullr(self, ix, iy, xs=None, ys=None):
        """ Get cell's canvas rectangle x, y  upper left, x,  y lower right
        
        :ix: cell x index
        :iy: cell's  y index
        :xs: grid x-values default: self.grid_xs
        :ys: grid y-values default: self.grid_ys 
        :returns: grid cell limits: (w_left_x,w_upper_y,
                                  w_right_x,w_lower_y) where
            w_left_x,w_upper_y: are upper left coordinates
            w_right_x,w_lower_y: are lower right coordinates
            Out of bounds - returns min/max in that direction
        """
        if xs is None:
            xs = self.grid_xs
        if ys is None:
            ys = self.grid_ys
        ixmax = len(xs)-2           # last element is upper bound
        iymax = len(ys)-2
        w_left_x = xs[0]            # left edge
        w_upper_y = ys[0]           # top edge
        w_right_x = xs[ixmax-1]
        w_lower_y = ys[iymax-1]
        
        if ix >= 0:
            if ix <= ixmax:
                w_left_x = xs[ix]
                w_right_x = xs[ix+1]
            else:
                w_left_x = xs[ixmax]
                w_right_x = xs[ixmax+1]
        if iy >= 0:
            if iy <= iymax:
                w_upper_y = ys[iy]
                w_lower_y = ys[iy+1]
            else:
                w_upper_y = ys[iymax]
                w_lower_y = ys[iymax+1]
        return (w_left_x,w_upper_y, w_right_x,w_lower_y)


    def get_canvas_items(self,
                        xmin=None, xmax=None, ymin=None, ymax=None,
                        ncols=None, nrows=None,
                        types=None, ex_types=None,
                        tags=None, ex_tags=None, get_color=False):
        """ Get items within grid cells
        :xmin,xmax,ymin,ymax, ncols, nrows: see get_grid_lims()
                        default: CanvasGrid instance values
        :types: list of types, select only items of these types default: All types
        :ex_types: List of types select only items not these types
                Only one of types or ex_types is allowed
        :tags: list of tags, include only items with one of these tags default: allow any tags
        :ex_tags: list of tags, exclude items with any of these tags default: no exclusions
                Only one of tags or ex_tags is allowed
        :get_color: True - returns list of cell colors default: return cell id
        :returns: list of overlapping canvas entries
                list entry format: tuple:
                                (ix,iy) - rectangle x index,  y index
                                    get_color==True: color
                                        list of entry info
                                    get_color==False: canvas item id
                                        color string
        """
                        # Get grid limites - defaulting to self.values
        xs,ys = self.get_grid_lims(xmin=xmin, xmax=xmax, ymin=ymin,ymax=ymax,
                                   ncols=ncols,nrows=nrows)
        ixy_ids_list = []       # Building list of (ix,iy), [overlapping ids]
        if types is not None and ex_types is not None:
            raise BrailleError("Don't support both types and ex_types")
         
        if tags is not None and ex_tags is not None:
            raise BrailleError("Don't support both tags and ex_tags")
         
        if types is not None:       # Support single or list of
            if not isinstance(types, list):
                types = [types]     # Make it a list
        if ex_types is not None:
            if not isinstance(ex_types, list):
                ex_types = [ex_types]     # Make it a list
        if tags is not None:
            if not isinstance(tags, list):
                tags = [tags]
        if ex_tags is not None:
            if not isinstance(ex_tags, list):
                ex_tags = [ex_tags]
        
        SlTrace.lg(f"get_canvas_items"
                   f" xmin={xmin}, xmax={xmax}, ymin={ymin}, ymax={ymax}", "get_items")
        for ix in range(len(xs)-1):
            for iy in range(len(ys)-1):
                cx1,cy1,cx2,cy2 = self.get_grid_ullr(ix=ix, iy=iy, xs=xs, ys=ys)
                item_ids_over_raw = self.base.find_overlapping(cx1,cy1,cx2,cy2)
                if SlTrace.trace("get_items"):
                    if len(item_ids_over_raw) > 0:
                        color = self.item_to_color(item_ids=item_ids_over_raw)
                    else:
                        color = ""
                    SlTrace.lg(f"    {ix},{iy}: cx1,cy1,cx2,cy2 {cx1},{cy1},{cx2},{cy2}"
                               f" item_ids: {item_ids_over_raw} {color}",
                               "get_items")
                if len(item_ids_over_raw) == 0:
                    continue    # Skip if none to check
                
                if (types is None and ex_types is None 
                        and tags is None and ex_tags is None):
                    item_ids_over = item_ids_over_raw     # Everything
                else:
                    item_ids_over = []
                    for item_id in item_ids_over_raw:
                        chosen = True 
                        if types:
                            itype = self.base.type(item_id)
                            if itype in types:
                                chosen = True    # Has required type
                            else:
                                chosen = False 
                        if ex_types:
                            itype = self.base.type(item_id)
                            if itype in ex_types:
                                chosen = False 
                            else:
                                chosen = True     
                        if chosen and tags:
                            chosen = False      # Set True if found
                            tag_list = self.base.gettags(item_id)
                            for tag in tag_list:
                                if tag in tags:
                                    chosen = True   # Has requisite flag 
                                    break
                        if chosen and ex_tags:
                            chosen = True      # Set False if found
                            tag_list = self.base.gettags(item_id)
                            for tag in tag_list:
                                if tag in ex_tags:
                                    chosen = False  # Has an excluded tag 
                                    break
                        if chosen:
                            item_ids_over.append(item_id)
                if len(item_ids_over) > 0:
                    if get_color:
                        color = self.item_to_color(item_ids=item_ids_over)
                        item_infos_over = color         # color string
                    else:
                        item_infos_over = item_ids_over # list of ids
                    ixy_ids_list.append(((ix,iy), item_infos_over))
        return ixy_ids_list

    def show_canvas(self, title=None, types=None, ex_types=None,
                  tags=None, ex_tags=None, get_color=False,
                  always_list=None,
                  xmin=None,ymin=None, xmax=None,ymax=None,
                  ncols=1,nrows=1,
                  prefix=None):
        """ Show whole canvas
            see show_canvas_items for args
        """
        if title is not None:
            SlTrace.lg(title)
        canvas_items = self.base.find_all()
        if len(canvas_items) == 0:
            SlTrace.lg("No canvas items")
            return
        
        for item_id in canvas_items:
            self.show_canvas_item(item_id=item_id, prefix=prefix)
        
    def show_canvas_items(self, title=None, types=None, ex_types=None,
                  tags=None, ex_tags=None, get_color=False,
                  always_list=None,
                  xmin=None,ymin=None, xmax=None,ymax=None,
                  ncols=None,nrows=None,
                  prefix=None):
        """ Show canvas items
        Args are those of get_canvas_items plus
        :title: title default: generated
        :always_list: list of options always listed
                    default: show_canvas_items default
        :prefix: prefix to output, used by show_canvas_item
        :xmin,...: default: self.xmin...
        """
        if title is not None:
            SlTrace.lg(title)
        ixy_ids_list = self.get_canvas_items(types=types,
                                ex_types=ex_types,
                                tags=tags, ex_tags=ex_tags,
                                get_color=get_color,
                                xmin=xmin,ymin=ymin, xmax=xmax,ymax=ymax,
                                ncols=ncols,nrows=nrows,
                                )
        xs,ys = self.get_grid_lims(xmin=xmin,ymin=ymin, xmax=xmax,ymax=ymax,
                                ncols=ncols,nrows=nrows)
        ixy_item_prefix = "" if prefix is None else prefix
        SlTrace.lg(f"{ixy_item_prefix} xs: {xs}")
        SlTrace.lg(f"{ixy_item_prefix} ys: {ys}")
        for ixy_item in ixy_ids_list:
            (ix,iy), ids = ixy_item
            ixy_item_prefix = "" if prefix is None else prefix
            w_left_x,w_upper_y, w_right_x,w_lower_y = self.get_grid_ullr(ix, iy,
                                                                     xs=xs,ys=ys)
            ixy_item_prefix += f" {ix},{iy}:"
            ixy_item_prefix +=  f" [{w_left_x},{w_upper_y}, {w_right_x},{w_lower_y}]"
            if len(ids) > 1:
                SlTrace.lg("")
            for item_id in ids:
                self.show_canvas_item(item_id=item_id, prefix=ixy_item_prefix,
                                      always_list=always_list)

                        
    def show_canvas_item(self, item_id, always_list=None,
                          prefix=None):
        """ display changing values for item
        :always_list: list of options which are always listed
                default: ['fill', 'width']
        """
        if prefix is None:
            prefix = ""
        if prefix != "" and not prefix.endswith(" "):
            prefix += " "       # leave some space
        if always_list is None:
            always_list = ['fill', 'width']  # Check always list
        iopts = self.base.itemconfig(item_id)
        itype = self.base.type(item_id)
        coords = self.base.coords(item_id)
        if itype in self.item_samples:
            item_sample_iopts = self.item_samples[itype]
        else:
            item_sample_iopts = None
     
        SlTrace.lg(f"{prefix} {item_id}: {itype} {coords}")
        ###self.show_coords(coords)
        
        for key in iopts:
            val = iopts[key]
            is_changed = True     # assume entry option changed
            if key not in always_list:  
                if item_sample_iopts is not None:
                    is_equal = True # Check for equal item option
                    sample_val = item_sample_iopts[key]
                    if len(val) == len(sample_val):
                        for i in range(len(val)):
                            if val[i] != sample_val[i]:
                                is_equal = False
                                break
                        if is_equal:
                            is_changed = False
            if is_changed:
                if isinstance(val, float):
                    SlTrace.lg(f"    {key} {val:.2f}")
                else:                     
                    SlTrace.lg(f"    {key} {val}")
            self.item_samples[itype] = iopts

    def create_magnify_info(self, x_min=None,y_min=None,
                    x_max=None,y_max=None,
                    ncols=None, nrows=None):
        """ Create a MagnifyInfo, using our values as defaults
        :x_min: minimum x value - left side 
        :y_min: minimum y value - top side
        :x_max: maximum x value - right side
        :y_max: maximum  y value - bottom side
        :ncols: number of grid columns
        :nrows: number of grid rows
        """
        if x_min is None:
            x_min = self.g_xmin
        if y_min is None:
            y_min = self.g_ymin
        if x_max is None:
            x_max = self.g_xmax
        if y_max is None:
            y_max = self.g_ymax
        if ncols is None:
            ncols = self.g_ncols
        if nrows is None:
            nrows = self.g_nrows
        
        top_region = MagnifyDisplayRegion(x_min=x_min, y_min=y_min,
                                          x_max=x_max, y_max=y_max,
                                          ncols=ncols, nrows=nrows)
        mag_info = MagnifyInfo(top_region=top_region,
                               base_canvas=self)
        return mag_info

    def get_mag_info_ullr(self, mag_info):
        """ Get boundaries from mag_info
        :mag_info: (MagnifyDisplayRegion)
        :returns: (left_x,upper_y,
                                    right_x,lower_y)
        """ 
        select = mag_info.select
        disp_region = mag_info.display_region
        if disp_region.ncols is None:
            disp_region.ncols = self.g_ncols
        if disp_region.nrows is None:
            disp_region.nrows = self.g_nrows
        disp_x_cell = (disp_region.x_max-disp_region.x_min)/disp_region.ncols
        disp_y_cell = (disp_region.y_max-disp_region.y_min)/disp_region.nrows
        xmin = select.ix_min*disp_x_cell + disp_region.x_min
        ymin = select.iy_min*disp_y_cell + disp_region.y_min
        xmax = (select.ix_max+1)*disp_x_cell + disp_region.x_min
        ymax = (select.iy_max+1)*disp_y_cell + disp_region.y_min
        return (xmin,ymin, xmax,ymax)   # left, top, right, bottom

    def show_mag_info_items(self, mag_info, ix, iy,
                            title=None, types=None, ex_types=None,
                  tags=None, ex_tags=None, get_color=False,
                  always_list=None,
                  prefix=None):
        """ Display canvas items for display
        :mag_info: (MagnifyDisplayRegion)
        :ix: ix index in mag_info
        :iy: iy index in mag_info
        """
        mag_info_copy = copy.copy(mag_info)     # Don't modify original
        mag_info_copy.select = MagnifySelect(ix_min=ix,ix_max=ix,
                                              iy_min=iy,iy_max=iy)
        xmin,ymin,xmax,ymax = self.get_mag_info_ullr(mag_info=mag_info_copy)
        self.show_canvas_items(title=title, types=types, ex_types=ex_types,
                  tags=tags, ex_tags=ex_tags, get_color=get_color,
                  always_list=always_list,
                  xmin=xmin,ymin=ymin, xmax=xmax,ymax=ymax,
                  ncols=1,nrows=1,
                  prefix=prefix)

                   
    def create_magnification_window(self, mag_info):
        """ Create magnification
        :mag_info: MagnificationInfo containing info
        :returns: instance of AudioDrawWinfow or None if none was created
        """
        select = mag_info.select
        disp_region = mag_info.display_region
        disp_x_cell = (disp_region.x_max-disp_region.x_min)/disp_region.ncols
        disp_y_cell = (disp_region.y_max-disp_region.y_min)/disp_region.nrows
        xmin = select.ix_min*disp_x_cell + disp_region.x_min
        ymin = select.iy_min*disp_y_cell + disp_region.y_min
        xmax = (select.ix_max+1)*disp_x_cell + disp_region.x_min
        ymax = (select.iy_max+1)*disp_y_cell + disp_region.y_min
        SlTrace.lg(f"create_magnification_window:"
                   f" xmin:{xmin} ymin:{ymin} xmax:{xmax} ymax:{ymax}"
                   f" nrows:{disp_region.nrows} ncols:{disp_region.ncols}")
        child_info = mag_info.make_child()
        child_info.display_region = MagnifyDisplayRegion(x_min=xmin, x_max=xmax,
                                        y_min=ymin, y_max=ymax)
        child_info.description = (f"region minimum x: {xmin:.0f}, minimum y: {ymin:.0f},"
                                  + f" maximum x: {xmax:.0f}, maximum y: {ymax:.0f}")
        adw = self.create_audio_window(xmin=xmin, xmax=xmax,
                                       ymin=ymin, ymax=ymax,
                                       nrows=disp_region.nrows,
                                       ncols=disp_region.ncols,
                                       mag_info=child_info,
                                       require_cells=True)            
        return adw 

    def exit(self):
        """ Main exit if creating magnifications
        """
        if self.pgmExit is not None:
            self.pgmExit()      # Use supplied pgmExit
            
        SlTrace.lg("CanvasGrid.exit")
        SlTrace.onexit()    # Force logging quit
        os._exit(0)

    def get_display_cells(self,
                 xmin=None, xmax=None, ymin=None, ymax=None,
                 nrows=None, ncols=None):
        """ Get cells to display, possibly in another process
        :xmin,xmax,ymin,ymax,: see get_grid_lims()
                        default: CanvasGrid instance values
        :nrows,ncols: grid size for scanning
                default: self.g_nrows, self.g_ncols
        :returns: list of tuples: (ix,iy,color_str)
        """
        if nrows is None:
            nrows = self.g_nrows
        if ncols is None:
            ncols = self.g_nrows    
        ixy_items = self.get_canvas_items(xmin=xmin, xmax=xmax,
                                          ymin=ymin,ymax=ymax,
                                          ncols=ncols,nrows=nrows)
        # For debugging / analysis
        xs,ys = self.get_grid_lims(xmin=xmin, xmax=xmax, ymin=ymin,ymax=ymax,
                                   ncols=ncols,nrows=nrows)
        braille_cells = []
        for ixy_item in ixy_items:
            (ix,iy), ids = ixy_item
            if SlTrace.trace("win_items"):
                SlTrace.lg(f"create_audio_window:{ix},{iy}: ids:{ids}")
                for canvas_id in ids:
                    self.show_canvas_item(canvas_id)
            color = self.item_to_color(item_ids=ids)
            if color is not None:
                cell = (ix, iy, color)
                braille_cells.append(cell)
        return braille_cells

    def create_audio_window(self, title=None,
                 xmin=None, xmax=None, ymin=None, ymax=None,
                 nrows=None, ncols=None, mag_info=None, pgmExit=None,
                 require_cells=False,
                 silent=False):
        """ Create AudioDrawWindow to navigate canvas from the section
        :title: optinal title
                region (xmin,ymin, xmax,ymax) with nrows, ncols
        :speaker_control: (SpeakerControlLocal) local access to centralized speech facility
        :xmin,xmax,ymin,ymax,: see get_grid_lims()
                        default: CanvasGrid instance values
        :nrows,ncols: grid size for scanning
                default: if mag_info present: mag_info.mag_nrows, .mag_ncols
        :mag_info: magnification info (MagnifyInfo)
                    default: None
        :pgm_exit: function to call upon exit request
        :require_cells: Require at least some display cells to 
                    create window
                    default: False - allow empty picture
        :silent: quiet mode default: False
        :returns: AudioDrawWindow instance or None if no cells
                Stores number of cells found in self.n_cells_created
        """
        from audio_draw_window import AudioDrawWindow
        from braille_cell import BrailleCell
                
        if pgmExit is None:
            pgmExit = self.exit 

        if mag_info is None:
            mag_info = self.create_magnify_info(x_min=xmin, y_min=ymin,
                                          x_max=xmax, y_max=ymax,
                                          ncols=ncols, nrows=nrows)
            
        if nrows is None:
            nrows = mag_info.mag_nrows
        if ncols is None:
            ncols = mag_info.mag_ncols    
        ixy_items = self.get_canvas_items(xmin=xmin, xmax=xmax,
                                          ymin=ymin,ymax=ymax,
                                          ncols=ncols,nrows=nrows)
        # For debugging / analysis
        xs,ys = self.get_grid_lims(xmin=xmin, xmax=xmax, ymin=ymin,ymax=ymax,
                                   ncols=ncols,nrows=nrows)
        braille_cells = []
        for ixy_item in ixy_items:
            (ix,iy), ids = ixy_item
            if SlTrace.trace("win_items"):
                SlTrace.lg(f"create_audio_window:{ix},{iy}: ids:{ids}")
                for canvas_id in ids:
                    self.show_canvas_item(canvas_id)
            color = self.item_to_color(item_ids=ids)
            if color is not None:
                bcell = BrailleCell(ix=ix, iy=iy, color=color)
                braille_cells.append(bcell)
        self.n_cells_created = len(braille_cells)
        if self.n_cells_created == 0:
            if require_cells:
                return None
        
        adw = AudioDrawWindow(title=title, speaker_control=None,
                              iy0_is_top=True, mag_info=mag_info, pgmExit=None,
                              x_min=self.g_xmin, y_min=self.g_ymin,
                              silent=silent)
        adw.draw_cells(cells=braille_cells)
        adw.key_goto()      # Might as well go to figure
        return adw

        
if __name__ == "__main__":
    import sys
    import time

    
    def test1():
        
        root = tk.Tk()
        cvg = TkCanvasGrid(root, height=800, width=800)
        for _ in range(2):
            cvg.paint_grid()
            time.sleep(2)
            cvg.erase_grid_paint()
            time.sleep(1)

    def test2():
        root = tk.Tk()
        cvg = TkCanvasGrid(root, height=450, width=450)
        cvg.create_line(0,0,200,300, width=10, fill="blue", tags="blue_tag")
        cvg.create_line(200,300,400,400, width=10, fill="green", tags=["green1","green2"])
        cvg.create_rectangle(200,200,300,300, fill="red")
        cvg.create_oval(100,200,250,300, fill="orange", tags="orange_tag")

        cvg.show_canvas_items(always_list=[])
        cvg.show_canvas_items(always_list=[],types="rectangle",prefix="rectangle")
        cvg.show_canvas_items(always_list=[],ex_types="rectangle",prefix="ex_types=rectangle ")
        cvg.show_canvas_items(always_list=[],ex_types="rectangle",
                              prefix="blue_tag, ex_types=rectangle ", tags="blue_tag")
        cvg.show_canvas_items(always_list=[],ex_types=["rectangle","line"],
                              prefix='ex_types=["rectangle","line"] ')
        cvg.show_canvas_items(always_list=[],ex_tags="green1",
                              prefix='ex_tags="green1"')

        SlTrace.lg("""Checking: tags="green1" when we have  tags=["green1","green2"] """)
        cvg.show_canvas_items(always_list=[],tags="green1",
                              prefix='tags="green1"')
        cvg.show_canvas_items(always_list=[],tags=["green1","orange_tag"],
                              prefix='tags=["green1","orange_tag"]')
        root.mainloop()
    
    def test3():    
        """ add cell squares to display
        """
        SlTrace.lg("Start test3")
        root = tk.Tk()
        cvg = TkCanvasGrid(root, height=450, width=450)
        cvg.create_line(0,0,200,300, width=10, fill="blue", tags="blue_tag")
        cvg.create_line(200,300,400,400, width=10, fill="green", tags=["green1","green2"])
        cvg.create_rectangle(200,200,300,300, fill="red")
        cvg.create_oval(100,200,250,300, fill="orange", tags="orange_tag")
        SlTrace.lg("calling display_cells()")
        cvg.display_cells()
        SlTrace.lg("After display_cells()")
        root.mainloop()
    
    def test4():    
        """ Create a AudioDrawWindow from a selection - default whole canvas
        """
        SlTrace.lg("Start test4")
        root = tk.Tk()
        cvg = TkCanvasGrid(root, height=450, width=450)
        cvg.create_line(0,0,200,300, width=10, fill="blue", tags="blue_tag")
        cvg.create_line(200,300,400,400, width=10, fill="green", tags=["green1","green2"])
        cvg.create_rectangle(200,200,300,300, fill="red")
        cvg.create_oval(150,250,350,350, fill="orange", tags="orange_tag")
        SlTrace.lg("Create a AudioDrawWindow")
        adw1 = cvg.create_audio_window()
        SlTrace.lg("After create_audio_window()")
        
        SlTrace.lg("Create a AudioDrawWindow 2")
        xmin = cvg.g_xmin + cvg.g_width//4
        xmax = cvg.g_xmax - cvg.g_width//4
        ymin = cvg.g_ymin + cvg.g_height//4
        ymax = cvg.g_ymax - cvg.g_height//4
        adw2 = cvg.create_audio_window(xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
        SlTrace.lg("After create_audio_window() 2")
        
        SlTrace.lg("Create a AudioDrawWindow 2")
        xmin = xmin+(xmax-xmin)*.3
        xmax = xmin+(xmax-xmin)*.7
        ymin = ymin+(ymax-ymin)*.3
        ymax = ymin+(ymax-ymin)*.7
        adw2 = cvg.create_audio_window(xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
        SlTrace.lg("After create_audio_window() 2")
        root.mainloop()
            
    def test4a():    
        """ Create a AudioDrawWindow from a selection - default whole canvas
        """
        SlTrace.lg("Start test4a")
        root = tk.Tk()
        cvg = TkCanvasGrid(root, height=450, width=450)
        cvg.create_line(0,0,200,300, width=10, fill="blue", tags="blue_tag")
        cvg.create_line(200,300,400,400, width=10, fill="green", tags=["green1","green2"])
        cvg.create_rectangle(200,200,300,300, fill="red")
        cvg.create_oval(150,250,350,350, fill="orange", tags="orange_tag")
        mag_info = MagnifyInfo(base_canvas=cvg)
        SlTrace.lg("Create a AudioDrawWindow")
        adw1 = cvg.create_audio_window()
        SlTrace.lg("After create_audio_window()")

        root.mainloop()
            
    def test4b():    
        """ Create a AudioDrawWindow from a selection - default whole canvas
        """
        SlTrace.lg("Start test4b")
        root = tk.Tk()
        height = 800
        width = height
        cvg = TkCanvasGrid(root, height=800, width=800)
        cvg.create_line(0,0,200,300, width=10, fill="blue", tags="blue_tag")
        cvg.create_line(200,300,400,400, width=10, fill="green", tags=["green1","green2"])
        cvg.create_rectangle(200,200,300,300, fill="red")
        cvg.create_oval(150,250,350,350, fill="orange", tags="orange_tag")
        #mag_info = MagnifyInfo(base_canvas=cvg)
        SlTrace.lg("Create a AudioDrawWindow")
        cvg.create_audio_window(xmin=0,ymin=0, xmax=width, ymax=height)
        SlTrace.lg("After create_audio_window()")

        root.mainloop()
    
    def test5():    
        """ Create a AudioDrawWindow from a selection - default whole canvas
        """
        SlTrace.lg("Start test5 using canvas scanning")
        root = tk.Tk()
        cv = tk.Canvas(root, height=450, width=450)
        cv.pack()
        cv.create_line(0,0,200,300, width=10, fill="blue", tags="blue_tag")
        cv.create_line(200,300,400,400, width=10, fill="green", tags=["green1","green2"])
        cv.create_rectangle(200,200,300,300, fill="red")
        cv.create_oval(150,250,350,350, fill="orange", tags="orange_tag")
        SlTrace.lg("Tk Canvas create")
        
        mw = tk.Tk()
        cvg = TkCanvasGrid(mw, base=cv, height=450, width=450)
        
        SlTrace.lg("Create a AudioDrawWindow")
        adw1 = cvg.create_audio_window(title="test5 from tk.canvas scanning")
        SlTrace.lg("After create_audio_window()")
        
        time.sleep(2)
        SlTrace.lg("Create a AudioDrawWindow 2")
        xmin = cvg.g_xmin + cvg.g_width//4
        xmax = cvg.g_xmax - cvg.g_width//4
        ymin = cvg.g_ymin + cvg.g_height//4
        ymax = cvg.g_ymax - cvg.g_height//4
        adw2 = cvg.create_audio_window(title="test5 Magnified window",
                                       xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
        SlTrace.lg("After create_audio_window() 2")
        
        time.sleep(1)
        SlTrace.lg("Create a AudioDrawWindow 3")
        xmin = xmin+(xmax-xmin)*.3
        xmax = xmin+(xmax-xmin)*.7
        ymin = ymin+(ymax-ymin)*.3
        ymax = ymin+(ymax-ymin)*.7
        adw2 = cvg.create_audio_window(title="test5 Magnified more window",
                                       xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
        SlTrace.lg("After create_audio_window() 2")
        root.mainloop()
    
    test = 5    
    if test == 1: test1()
    if test == 2: test2()
    if test == 3: test3()
    if test == "4a": test4a()
    if test == "4b": test4b()
    if test == 5: test5()
    SlTrace.lg("End of Test")
    sys.exit()
        


# braille_cell_text.py    11Apr2023  crs, Author
"""
BrailleCell group to text picture
Text picture to BrailleCell group
For testing / Diagnostics
Plan:
Build BrailleCell dictionary/list from text picture
Text picture examples:
"""
import re 

from select_trace import SlTrace
from braille_cell import BrailleCell

yellow_circle ="""
,,,,,,,,,,,,yy
,,,,,,,,,,yyyyy
,,,,,,,,,,yyyyy
,,,,,,,,,,yyyyy
,,,,,,,,,,,yyy
"""
yellow_circle_modified ="""
,,,,,,,,,,,,yy
,,,,,,,,,,yyyyy
,,,,,,,,,,ygivy
,,,,,,,,,,yyyyy
,,,,,,,,,,,yyy
"""

spokes_picture="""
,,,,,,,,,,,iii
,,,,,,,,,,iiiii
,,,,,,,,,,iiiii,,,,,,vvv
,,,,,,,,,,iiiii,,,,,vvvvv
,,,,,,,,,,,,ii,,,,,,vvvvv
,,,bb,,,,,,,,i,,,,,,vvvvv
,,bbbbb,,,,,,i,,,,,vv
,,bbbbb,,,,,,i,,,,vv
,,bbbbbbb,,,,ii,,vv
,,,,,,,,bbbb,,i,vv,,,,,,,,rr
,,,,,,,,,,bbbbivv,,,,,,,,rrrr
,,,,,,,,,,,,,bvvrrrrrrrrrrrrr
,,,,,,,,,,ggggyoo,,,,,,,,rrrr
,,,,,,,,gggg,,y,oo,,,,,,,,rr
,,ggggggg,,,,yy,,oo
,,ggggg,,,,,,y,,,,oo
,,ggggg,,,,,,y,,,,,oo
,,,gg,,,,,,,,y,,,,,,ooooo
,,,,,,,,,,,,yy,,,,,,ooooo
,,,,,,,,,,yyyyy,,,,,ooooo
,,,,,,,,,,yyyyy,,,,,,ooo
,,,,,,,,,,yyyyy
,,,,,,,,,,,yyy
"""

class BrailleCellText:
    """ translator between Cells and text picture
    """
    def __init__(self, text=None, cells=None,
                 grid_height=25, grid_width=32,
                 cell_nch=1):
        """ Setup text/cells, given the other
        :text: text string portraying the cells
                either "," or " " (space) is supported
                for input
        :cells: dictionary by ix,iy or list of BrailleCell
        Only one of text or cells is allowed_attributes
        :grid_height: assumed window height default: 25
        :grid_width: assumed window width default: 32
        :cell_nch: number of chars per cell default:1
        """
        if text is not None and cells is not None:
            SlTrace.lg("Only one of text or cells may be specified")
            return
        
        self.text = text
        self.cells = cells
        self.grid_height = grid_height
        self.grid_width = grid_width
        self.cell_nch = cell_nch
        
        if text is not None:
            self.build_from_text(self.text)
        else:
            self.build_from_cells(cells=self.cells) 


    def build_from_cells(self, cells=None,
                         shift_to_edge=True, blank_char=",",
                         cell_nch=None):
        """ Build text from cells
        :cells: list/dictionary by ix,iy cells to build text
            default: use self.cells
        :shift_to_edge: shift figure towards edge to ease finding figure
                        default: self.shift_to_edge
        :blank_char: convert leading blanks to this
                    default: ","
        :cell_nch:    Number of text chars per cell
                    default: from BrailleCellText init
        """
        if cell_nch is None:
            cell_nch = self.cell_nch
        if cells is None:
            cells = self.cells
        if isinstance(cells, list):
            cdict = {}
            for cell in cells:
                cdict[(cell.ix,cell.iy)] = cell
            cells = cdict
        self.cells = cells
        self.text = self.get_text(shift_to_edge=shift_to_edge,
                                  blank_char=blank_char,
                                  cell_nch=cell_nch)
        
    def get_text(self, cells=None, shift_to_edge=None, blank_char= ",",
                 marker_ixy=None, marker=None, marker_space=None,
                 cell_nch=None):
        """ get text picture for braille display from cells
        :shift_to_edge: shift figure towards edge to ease finding figure
                        default: self.shift_to_edge
        :blank_char: convert leading blanks to this
                    default: ","
        :marker_ixy: ixy tuple to mark default: No marking
        :marker: Type marking
                "uppercase" - uppercase the expected character
                ONE char - e.g. X - use this character
                default: "uppercase"
        :marker_space: Marker for space default: "."
        :cell_nch:    Number of text chars per cell
                    default: from __init__
        """
        if cell_nch is None:
            cell_nch = self.cell_nch
        if cells is None:
            cells = self.cells
        if shift_to_edge is None:
            shift_to_edge = True
        self.blank_char = blank_char
        self.find_edges(cells=cells)
        left_edge = self.left_edge
        right_edge = self.right_edge
        top_edge = self.top_edge
        bottom_edge = self.bottom_edge
        
        if not shift_to_edge:
            left_edge = 0
            top_edge = 0

        braille_text = ""
        for iy in range(top_edge, bottom_edge+1):
            line = ""
            for ix in range(left_edge, right_edge+1):
                cell_ixy = (ix,iy)
                if cell_ixy in self.cells:
                    cell = self.cells[cell_ixy]
                    color = cell.color_string()
                    if cell.pi_number is not None:
                        mark = str(cell.pi_number)
                        if len(mark) < cell_nch:
                            mark = color[0] + mark
                        while len(mark) < cell_nch:
                            mark += color[0]
                    else:
                        mark = color[0]
                        while len(mark) < cell_nch:
                            mark += color[0]
                    if marker_ixy == cell_ixy:
                        if marker == "uppercase" or marker is None:
                            mark = mark.upper()
                        else:
                            mark = f"[{mark}]" 
                else:
                    mark = " "
                    if  marker_ixy == cell_ixy:
                        mark = marker_space
                if len(mark) == 1 and cell_nch > 1:
                    mark *= cell_nch
                while len(mark) < cell_nch:
                    mark += "_"
                line += mark
            line = line.rstrip()
            if self.blank_char != " ":
                line = line.replace(" ", self.blank_char)
            ###print(f"{iy:2}", end=":")
            if not re.match(r"^\s*$", line):
                braille_text += line + "\n"
        return braille_text
                    
    def build_from_text(self, text):
        cells = {}
        ix = 0
        iy = 0
        for ch in text:
            if ch == "," or ch == " ":
                ix += 1             
            elif ch in BrailleCell.dots_for_character:
                color = BrailleCell.ch2color(ch=ch)
                cell = BrailleCell(ix=ix, iy=iy, color=color)
                cells[(ix,iy)] = cell
                ix += 1
            elif ch == "\n":
                ix = 0
                iy += 1
            else:
                ix += 1     # Ignore but keep alignment
        self.cells = cells
                 
    def print_text(self, shift_to_edge=None, blank_char= ",",
                 marker_ixy=None, marker=None, marker_space=None,
                 cell_nch=None):
        """ Print out (to log) text rendition
        :shift_to_edge: shift figure towards edge to ease finding figure
                        default: self.shift_to_edge
        :blank_char: convert leading blanks to this
                    default: ","
        :marker_ixy: ixy tuple to mark default: No marking
        :marker: Type marking
                "uppercase" - uppercase the expected character
                ONE char - e.g. X - use this character
                default: "uppercase"
        :marker_space: Marker for space default: "."
        :cell_nch: cell number of characters
                    default: self.cell_nch
        """
        if cell_nch is None:
            cell_nch = cell_nch
        SlTrace.lg()
        text = self.get_text(shift_to_edge=shift_to_edge,
                              blank_char= blank_char,
                              marker_ixy=marker_ixy, marker=marker,
                              marker_space=marker_space,
                              cell_nch=cell_nch)
        SlTrace.lg(text)

    def diff_text(self, text, space_comma=None, trim_spaces=True,
                  shift_to_edge=True,
                  blank_char= ",",
                  cell_nch=None):
        """ Return difference in text line by line
        :space_comma: space and comma replaced with this
                        default: " "  None -> no replacement
        :trim_spaces: trim leading/trailing whitespace before comparison
                        default: True
                        Done after any space_comma replacement
        :returns: lines with differences
        :shift_to_edge: shift figure towards edge to ease finding figure
                        default: self.shift_to_edge
        :blank_char: convert leading blanks to this
                    default: ","
        :cell_nch:    Number of text chars per cell
                    default: from __init__
        """
        if cell_nch is None:
            cell_nch = self.cell_nch
        if space_comma is None:
            space_comma = blank_char
        our_text = self.get_text(shift_to_edge=shift_to_edge,
                                 blank_char= blank_char,
                                 cell_nch=cell_nch)
        our_lines = our_text.splitlines()
        if shift_to_edge:
            text_bct = BrailleCellText(text=text, cell_nch=cell_nch)
            text = text_bct.get_text(shift_to_edge=shift_to_edge)
        text_lines = text.splitlines()
        diff = ""
        while len(our_lines)>0 and our_lines[0] == '':
            our_lines.pop(0)
        while len(our_lines)>0 and our_lines[-1] == '':
            our_lines.pop()
        while len(text_lines)>0 and text_lines[0] == '':
            text_lines.pop(0)
        while len(text_lines)>0 and text_lines[-1] == '':
            text_lines.pop()

        while len(our_lines) >0 or len(text_lines) > 0:
            our_line = "" if len(our_lines)==0 else our_lines.pop(0)
            text_line = "" if len(text_lines)==0 else text_lines.pop(0)
            if space_comma is not None:
                our_line = our_line.replace(',', space_comma)
                our_line = our_line.replace(' ', space_comma)
                text_line = text_line.replace(',', space_comma)
                text_line = text_line.replace(' ', space_comma)
            if trim_spaces:
                our_line = our_line.strip()
                text_line = text_line.strip()
            if our_line != text_line:
                our_diff  = " < " + our_line + "\n"
                text_diff = " > " + text_line + "\n"
                diff += our_diff + text_diff
        return diff
    
    def get_cells(self):
        """ Returns cells(dictionary by ixy) created
        """
        return self.cells
            
    def print_cells(self):
        """ Print our cells a line at a time
        """
        ix_max = self.get_ix_max()
        iy_max = self.get_iy_max()
        for iy in range(iy_max+1):
            st_line = ""
            for ix in range(ix_max+1):
                ixy = (ix,iy)
                if ixy in self.cells:
                    st_line += f" {self.cells[ixy]}"
            if st_line != "":
                SlTrace.lg(st_line)

    """
         From audio_draw_window
         with the ix_min.... adjusted for local
    """
    def get_ix_min(self):
        """ get minimum ix on grid
        :returns: min ix
        """
        return 0

    def get_ix_max(self):
        """ get maximum ix on grid
        :returns: min ix
        """
        return self.grid_width-1

    def get_iy_min(self):
        """ get minimum iy on grid
        :returns: min iy
        """
        return 0

    def get_iy_max(self):
        """ get maximum ix on grid
        :returns: min ix
        """
        return self.grid_height-1

    def find_edges(self, cells=None):
        """Find  top, left, bottom, right non-blank edges
        so we can shift picture to left,top for easier
        recognition
        :cells: cells default:self.cells
        """
        if cells is None:
            cells = self.cells
        left_edge,top_edge, right_edge,bottom_edge = self.bounding_box_ci(
                                cells=cells)
        if bottom_edge is None:
            bottom_edge = self.grid_height-1
                    
        if left_edge > 0:           # Give some space
            left_edge -= 1
        if left_edge > 0:
            left_edge -= 1
        self.left_edge = left_edge
        
        if top_edge > 0:           # Give some space
            top_edge -= 1
        if bottom_edge  < self.grid_height-1:
            bottom_edge += 1        
        self.top_edge = top_edge
        self.bottom_edge = bottom_edge
        
        self.left_edge = left_edge
        self.right_edge = right_edge
         
        return left_edge, top_edge, right_edge, bottom_edge
    
    def bounding_box_ci(self, cells=None, add_edge=None):
        """ cell indexes which bound the list of cells
        :cells: list of cells, (with cell.ix,cell.iy) or (ix,iy) tuples
                default: list of all cells in figure
        :add_edge: number of cells to add/subtract (if possible)
                     to enlarge/shrink box
                    default: no change
        :returns: 
                    None,None,None,None if no figure
                    upper left ix,iy  lower right ix,iy
        """
        if cells is None:
            if not hasattr(self, "cells"):
                return None,None,None,None         # Not yet setup
            else:
                cells = self.cells
        
        ix_min, iy_max, ix_max,iy_min = None,None,None,None
        for cell in cells:
            cell_ixy = cell if isinstance(cell, tuple) else (cell.ix,cell.iy)
            cell_ix, cell_iy = cell_ixy
            if ix_min is None or cell_ix < ix_min:
                ix_min = cell_ix
            if ix_max is None or cell_ix > ix_max:
                ix_max = cell_ix
            if iy_min is None or cell_iy < iy_min:
                iy_min = cell_iy
            if iy_max is None or cell_iy > iy_max:
                iy_max = cell_iy

        if ix_min is None:
            ix_min = 0
        if ix_max is None:
            ix_max = self.grid_width-1
        if iy_min is None:
            iy_min = 0
        if iy_max is None:
            ix_max = self.grid_height-1
        
        if add_edge is not None:    # Extend/Shrink box
            ext_ix_min = ix_min - add_edge
            bd_ix_max = self.get_ix_max() # protect against too large of negative add_edge
            if ext_ix_min > bd_ix_max: ext_ix_min = bd_ix_max
            ix_min = max(ext_ix_min, self.get_ix_min()) # limit to bounds

            ext_iy_min = iy_min - add_edge
            bd_iy_max = self.get_iy_max()
            if ext_iy_min > bd_iy_max: ext_iy_min = bd_iy_max
            iy_min = max(ext_iy_min, self.get_iy_min())
            
            ext_ix_max = ix_max + add_edge
            bd_ix_min = self.get_ix_min()
            if ext_ix_max < bd_ix_min: ext_ix_max = bd_ix_min
            ix_max = min(ext_ix_max, self.get_ix_max())
            
            ext_iy_max = iy_max + add_edge
            bd_iy_min = self.get_iy_min()
            if ext_iy_max < bd_iy_min: ext_ix_max = bd_iy_min
            iy_max = min(ext_iy_max, self.get_iy_max())
            
        return ix_min,iy_min, ix_max,iy_max

        
        
if __name__ == "__main__":
    SlTrace.clearFlags()
    SlTrace.lg("\n\nTest: yellow_circle")
    test_text = yellow_circle
    ycir_bct = BrailleCellText(text=test_text)
    SlTrace.lg(f"{test_text} cells:")
    ycir_bct.print_cells()
    SlTrace.lg(f"{test_text} unmodified text:")
    ycir_bct.print_text(shift_to_edge=False)

    SlTrace.lg("Testing text diff")
    SlTrace.lg(f"yellow_circle_modified:")
    SlTrace.lg(yellow_circle_modified)
    diff = ycir_bct.diff_text(yellow_circle_modified)
    for line in diff.splitlines():
        SlTrace.lg(f"diff: {line}")
    
    SlTrace.lg("\n\nTest: spokes_picture")
    test_text = spokes_picture
    spokes_bct = BrailleCellText(text=test_text)
    SlTrace.lg("to cells:")
    spokes_bct.print_cells()
    SlTrace.lg("From text:")
    spokes_bct.print_text(shift_to_edge=False)

    chg_ix, chg_iy = 5,10
    chg_ixy = (chg_ix,chg_iy)
    chg_color = "w"
    SlTrace.lg(f"Changing {chg_ixy} color to {chg_color}")
    if chg_ixy in spokes_bct.cells:
        cell = spokes_bct.cells[chg_ixy]
        SlTrace.lg(f"Color:{cell.color.string()} to {chg_color}")
    else:
        SlTrace.lg(f"New cell at {chg_ixy} color={chg_color}")
        cell = BrailleCell(ix=chg_ix, iy=chg_iy, color=chg_color)
        spokes_bct.cells[chg_ixy] = cell
    SlTrace.lg()
    SlTrace.lg("Converting cells back to text")
    bct2 = BrailleCellText(cells=spokes_bct.get_cells())
    SlTrace.lg("Comparing with original text")
    diff_spokes = bct2.diff_text(spokes_picture)
    for line in diff_spokes.splitlines():
        SlTrace.lg(f"diff: {line}")
    
    SlTrace.lg("Printing resulting text")
    bct2.print_text()
    
    cell_nch = 4
    SlTrace.lg(f"Displaying text with cell_nch = {cell_nch}")
    spokes_bct.print_text(cell_nch=4)


#wx_key_cmd_proc.py 30Jan2024  crs, Author
""" Process key commands
for AutioDrawWindow display
Provids a delay between commands to emulate physical
behavior.
The commands are placed in a queue which is emptied at
a suitable interval.  A separte thread is used to
allow the wx.App processing of events to support
display.
"""
import time
import threading as th
import queue
import wx

from select_trace import SlTrace


class KeyCmdProc:
    def __init__(self, win, key_press, dly=.2):
        """ Setup keyboard cmd processing
        :adw: AudioDrawWindow
        :dly: delay between actions in seconds
                default: .2
        """
        self.key_press = key_press  # key function to call
        self.dly = dly
        self._running = True # processing
        self._doing_command = False     # Not processing cmd
        self._awaiting = False  # True - awaiting callback 
        self.cmd_queue = queue.Queue()  # Command queue
        self.cmd_thread = th.Thread(target=self.thread_process)
        self.cmd_thread.start()
        
    def thread_process(self):
        """ Process cmds
        """
        while self._running:
            if not self.cmd_queue.empty():
                wx.CallAfter(self.delay_next_call)
            time.sleep(1.)
            
    def do_next_command(self):
        """ Process next command from queue, if one
        """
        SlTrace.lg("do_next_command", "cmd")
        self._doing_command = True
        self._awaiting = False
        if self.cmd_queue.empty():
            self._doing_command = False
            return      # Nothing to do
    
        cmd = self.cmd_queue.get()
        self.key_press(cmd)
        self._awaiting = True
        self._doing_command = False
        wx.CallAfter(self.delay_next_call)
    
    def delay_next_call(self):
        SlTrace.lg("delay_next_call", "cmd")
        wx.CallLater(int(self.dly*1000), self.do_next_command)

    def empty(self):
        """ Test if empty (no pending cmds)
        :returns: True iff queue is empty
        """
        return self.cmd_queue.empty()
    
    def stop(self, wait=True):
        """ Stop processing, end thread
        :wait: True wait till empty
                default: True
        """
        if wait:
            wx.CallAfter(self.stop_on_empty)
        else:
            self._running = False

    def stop_on_empty(self):
        """ Stop when empty
        """
        if self.cmd_queue.empty():
            self._running = False
            return
        
        wx.CallLater(100, self.stop_on_empty) 

    def put_key_cmd(self, cmd):
        """Place key cmd in queue to be processed
        """
        self.cmd_queue.put(cmd)
        if not self._awaiting and not self._doing_command:
            self.do_next_command()
            
if __name__ == '__main__':
    import time
    import wx
    
    app = wx.App()
    win = wx.Frame(None)
    win.Show()
    
    def test_1():
        print("Test Start")
            
        def key_press(ch):
            SlTrace.lg(f"key_press({ch})", "cmd_out")
            print(ch)
            
        kcp = KeyCmdProc(win, key_press=key_press, dly=.200)
        
        def key_str(key_str):
            """ Process string via kcp
            :key_str: character string to process
            """
            for ch in key_str:
                kcp.put_key_cmd(ch)

        ab = "abcdefghijklmnopqrstuvwxyz"
        ab = ab[:5]
        chs = ab + ab.upper() 
        key_str(chs)
        chs = chs[::-1]
        key_str(chs)
        wx.CallAfter(print, f"after key_str({chs})")
    
    #SlTrace.setFlags("cmd_out")    
    wx.CallAfter(test_1)
    
    app.MainLoop()


# wx_speaker_control.py    24Oct2023  crs, Author
"""
Support thread safe non-blocking speaker control:
    1. text to speech encapsulation of pyttsxN facilitating
     talk from multiple AudioDrawWindow sources
    2. tone playing using SineWaveNumPy, audio-sounddevice
    3. Using wxPython utilities for time control
    4. Using PyttsxProc for speech to text
    
"""
import threading
import queue
import time
import wx

import sounddevice as sd

from format_exception import format_exception
from select_trace import SlTrace, SelectError
from wx_play_sound_control import PlaySoundControl
from sinewave_numpy import SineWaveNumPy

from pyttsx_proc import PyttsxProc
from speech_maker_cmd import SpeechMakerCmd



class SpeakerControlError(SelectError):
    pass

class SpeakerControlDelay:
    """ Delay info
    """
    def __init__(self, sc, dur):
        """ Start delay
        :sc: speaker control instance
        :dur: duration seconds
        """
        self.sc = sc
        self.dur = dur
        self.start = time.time()
        self.end = self.start + self.dur
        self.waiting = True 
        
    def is_end(self):
        """ Check if time is up
        :returns: True if at or past end
        """
        now = time.time()
        if not self.waiting or now > self.end:
            return True 
        
        return False

    def stop(self):
        """ Stop delay (set as over)
        """
        self.waiting = False

class Singleton:
    _instance = None
    _lock = threading.Lock()

    def __new__(cls):
        if cls._instance is None: 
            with cls._lock:
                # Another thread could have created the instance
                # before we acquired the lock. So check that the
                # instance is still nonexistent.
                if not cls._instance:
                    cls._instance = super().__new__(cls)
        return cls._instance


class SpeakerTone:
    """ Tone to emit
    """
    
    def __init__(self, pitch, volume, dur, delay=0):
        """ Setup tone to emit
        :pitch:
        :volume: (v_left, v_right)
        :dur: seconds
        :delay:   delay before start default: no delay
        """
        self.pitch = pitch
        if volume is None:
            SlTrace.lg("SpeakerTone volume={volume}, treat as 0,0")
            volume = (0,0)
        self.volume = volume
        if dur is None:
            SlTrace.lg(f"SpeakerTone dur is {dur} treat as .1")
            dur = .1
        self.dur = dur
        self.delay = delay
    
    def __str__(self):
        st = f"SpeakerTone: pitch={self.pitch}"
        if self.volume is not None:
            v_left,v_right = self.volume
            st += f" vol=({v_left:.1f},{v_right:.1f})"
        if self.dur is None:
            st += " dur=None"
        else:
            st +=  f" dur={self.dur:.3f}"
        if self.delay is not None:
            st += f" delay:{self.delay:.3f}"
        return st

class SpeakerWaveform:
    """ waveform to emit
    """
    
    def __init__(self, ndarr, dur,
                 calculate_dur=None, delay=None, sample_rate=None):
        """ setup waveform
        :ndarr: waveform (SinewaveNumPy stereo_waveform)
        :dur: wave max duration (seconds)
        :calculate_dur: calculate duration based on waveform and sample_rate
        :dly: delay(seconds) from start default: no delay
        :sample_rate: sample rate fps
        """
        self.ndarr = ndarr
        self.dur = dur
        self.calculate_dur = calculate_dur
        self.delay = delay
        self.sample_rate = sample_rate
    
    def __str__(self):
        st = f"SpeakerWaveform: dur={self.dur}"
        if self.delay is not None:
            st += f" delay:{self.delay:.3f}"
        return st
    
class SpeakerControlCmd:
    """ Command to execute
    """

    def __init__(self, cmd_type=None, msg=None, msg_type=None,
                 rate=None, volume=None,
                 tone=None, waveform=None, fr=None, after=None):
        """ Setup command
        :cmd_type: command to execute
                "CMD_MSG" - speak message - msg REQUIRED
                "CMD_SPEAK_TEXT_STOP" - stop current/pending text speech
                "CMD_TONE" - make tone - tone REQUIRED
                "CMD_WAVEFORM - make wf waveform REQUIRED
                "CLEAR" - clear pending speech/tone
                "QUIT" - quit operation
        :msg: text to speak SpeakText
        :msg_type: type of message
                REPORT: std reporting
                CMD: command
                ECHO: echo of user input
            default: REPORT
        :rate: speach rate WPM
        :volume: speach volume
        :tone: (SpeakerTone) tone to make
        :waveform: (SpeakerWaveform) waveform to play 
        :fr: SpeakerControlLocal's window reference
        :after: function to call after cmd completion
                default: no call
        """
        self.cmd_id = SpeakerControl.new_id()
        self.cmd_type = cmd_type
                
        self.msg = msg
        self.msg_type = msg_type
        self.rate = rate
        self.volume = volume
        self.tone = tone
        self.waveform = waveform
        self.fr = fr
        self.after = after
        if after is not None:
            if fr is None:
                ###wxport###raise SpeakerControlError(f"after missing fr cmd:{self}")
                pass
        if cmd_type == "CMD_MSG":
            if msg is None:
                raise SpeakerControlError(f"msg missing with type MSG: {self}")
        if cmd_type == "CMD_TONE":
            if tone is None:
                raise SpeakerControlError(f"tone missing with type TONE: {self}")
        if cmd_type == "CMD_WAVEFORM":
            if waveform is None:
                raise SpeakerControlError(f"waveform missing with type WF: {self}")
        

    def __str__(self):
        ret = f"SpeakerControlCmd: {self.cmd_id} {self.cmd_type}"
        if self.msg is not None:
            ret += f" {self.msg}"
        if self.msg_type is not None:
            ret += f" {self.msg_type}"
        if self.tone is not None:
            ret += f" {self.tone}"
        if self.waveform is not None:
            ret += f" {self.waveform}"
        if self.after is not None:
            ret += f" fr:{self.fr} after:{self.after}"
        
        return ret

        
class SpeakerControl(Singleton):
    CMDS_SIZE = 150
    SOUND_SIZE = 150
    #SOUND_SIZE = 30        # To force filling
    cmd_id = 0
    
    @classmethod
    def new_id(cls):        # Thread safe ???  TBD
        cls.cmd_id += 1
        return cls.cmd_id
    
    def __init__(self, cmds_size=CMDS_SIZE, sound_size=SOUND_SIZE,
                 sample_rate=44100):
        """ Setup speaker control
        :cmds_size: general command queue size
        :sounc_size: pending sounds queue size
        :sample_rate: waveform presentation sample_rate
                    default: 44100 per second
        :simple_speaker: For Debugging/Test
                True - avoid multiprocessing/threading to aboid freeze problems
                default: False
        """
        self.pyttsx_proc  = PyttsxProc()
        self.cmds_size = cmds_size
        self.sound_size = sound_size
        self.sample_rate = sample_rate
        self.vol_adj = 0.0      # Central volume adjustment factor
        self.start_control()
        
    def start_control(self):
        """ Start / Restart control
        """
        self._running = True        # Thread functions exit when cleared
        self.sc_busy = False        # True -> control busy
                                    # to replace other busys
        self.forced_clear = False   # set on force_clear, checked on waiting...
        self.sound_busy = False     # Set True if active or pending

        self.psc = PlaySoundControl()
        self.sound_lock = threading.Lock()
        self.cmds_in_progress = {}   # cmd ins process by cmd_id
        self.sc_cmd_queue = queue.Queue(self.cmds_size)  # Command queue of SpeakerControlCmd
        self.sc_cmd_thread = threading.Thread(target=self.sc_cmd_proc_thread)
        self.sc_sound_queue = queue.Queue(self.sound_size)  # speech queue of SpeakerControlCmd 
        self.sc_sound_thread = threading.Thread(target=self.sc_sound_proc_thread)
        self.sc_sound_thread.start()
        self.sc_cmd_thread.start()
        self.sc_tone_busy = False       # True - iff toneing

        
    def sc_cmd_proc_thread(self):
        """ speech maker command processing thread function
        """
        while not self.forced_clear and self._running:            
            cmd = self.sc_cmd_queue.get()
            SlTrace.lg(f"cmd: {cmd}", "speech")
            SlTrace.lg(f"cmd: {cmd}", "sound_queue")
            if cmd.cmd_type == "CMD_MSG":
                SlTrace.lg(f"CMD_MSG: {cmd}", "CMD_MSG")
                self.sc_sound_queue.put(cmd)
            elif cmd.cmd_type == "CMD_STOP_SPEAK_TEXT":
                self.stop_speak_text()            
            elif cmd.cmd_type == "CMD_TONE":
                self.sc_sound_queue.put(cmd)
            elif cmd.cmd_type == "CMD_WAVEFORM":
                self.sc_sound_queue.put(cmd)
            elif cmd.cmd_type == "CLEAR":
                SlTrace.lg("Clearing speech")
                self.clear()
            elif cmd.cmd_type == "QUIT":
                self.clear()
                self.sc_sound_queue.put(cmd)
                self.quit()
            else:
                raise SpeakerControlError(f"Unrecognized SpeakerControl command {cmd}")
        SlTrace.lg("sc_cmd_proc_thread returning")

    def get_cmd_queue_size(self):
        """ Get current number of entries
        :returns: number of entries
        """
        return self.sc_cmd_queue.qsize()

    def get_sound_queue_max(self):
        """ Get current number of entries
        :returns: number of entries
        """
        return self.self.sound_size

    def get_sound_queue_size(self):
        """ Get current number of entries
        :returns: number of entries
        """
        return self.sc_sound_queue.qsize()

    def get_vol_adj(self):
        """ Get current volume adjustment
        :returns: current vol_adjustment in db
        """
        return self.vol_adj

    def report_vol_adj(self):
        SlTrace.lg(f"vol_adj: {self.vol_adj}")

    def set_vol_adj(self, adj=0.0):
        """ Set volume adjustment
        :adj: db adjustment default:0.0
        """
        self.vol_adj = adj
        self.report_vol_adj()
        
    def clear(self):
        """ Clear queue
        """
        SlTrace.lg("speech clearing")
        sd.stop()
        self.clear_cmd_queue()
        SlTrace.lg(f"self.sc_cmd_queue.qsize(): {self.sc_cmd_queue.qsize()}")
        self.clear_sound_queue()
        SlTrace.lg(f"self.sc_sound_queue.qsize(): {self.sc_sound_queue.qsize()}")
        self.sc_speech_busy = False 
        self.sc_tone_busy = False
        self.sound_busy = False
    
    def stop_scan(self):
        """ Stop current scan
        """
        self.clear()
    
    def stop_speak_text(self):
        """ Stop current and pending text speach
        """
        self.pyttsx_proc.clear()
        

    def clear_cmd_queue(self):
        while self.sc_cmd_queue.qsize() > 0:
            cmd = self.sc_cmd_queue.get()
            SlTrace.lg(f"removing cmd entry: {cmd}")

    def clear_sound_queue(self):
        while self.sc_sound_queue.qsize() > 0:
            cmd = self.sc_sound_queue.get()
            SlTrace.lg(f"removing speech queue entry: {cmd}")
            
    def force_clear(self):
        """ force Clear
        :restart: restart controller after a short wait
        """
        SlTrace.lg("force speech clearing")
        self.clear_cmd_queue()
        SlTrace.lg(f"self.sc_cmd_queue.qsize(): {self.sc_cmd_queue.qsize()}")
        self.clear_sound_queue()
        SlTrace.lg(f"self.sc_sound_queue.qsize(): {self.sc_sound_queue.qsize()}")
        sd.stop()                   # Stop waveform
        self.forced_clear = True        # stoping waits...

        #if self.pyttsxN_engine.isBusy():
        #self.pyttsxN_engine.stop()
        #if self.pyttsxN_engine._inLoop:
        #    self.pyttsxN_engine.endLoop()
        #self.pyttsxN_engine = pyttsxN.init()

        #self.sound_lock.release()
        self.sc_speech_busy = False 
        self.sc_tone_busy = False

    def force_clear_reset(self):
        self.sc_forced_clear = False 

            
    def busy_parts(self):
        """ Check if if busy parts
        """
        if self.pyttsx_proc.is_busy() or self.sc_tone_busy:     # Fast check
            return True
        
        if self.get_cmd_queue_size() > 0:
            return True
         
        if self.get_sound_queue_size() > 0:
            return True
        
        return False

            
    def is_busy(self):
        """ Check if if busy or anything is pending
        """
        if self.sc_busy or self.busy_parts():
            self.sc_busy = self.busy_parts()
            return True
                
        return False

    def is_in_progress(self, cmd_id):
        """ Check if cmd is still in progress
        :cmd_id: command  id
        """
        if cmd_id in self.cmds_in_progress:
            return True 
        
        return False 
    
    def sc_sound_proc_thread(self):
        """ Process pending speech requests (SpeakerControlCmd)
        """
        SlTrace.lg("sc_sounc_proc_thread running", "sound")
        #while not self.forced_clear and self._running:
        while True:
            qsize = self.sc_sound_queue.qsize()
            SlTrace.lg(f"sound_queue {qsize}: BEFORE sc_sound_queue.get()", "sound_queue")
            buffer = 10
            space_left = self.sound_size - qsize
            if space_left < buffer:
                SlTrace.lg(f"Speech queue filling qsize:{qsize}"
                           f" of {self.sound_size}")
                while True:
                    qsize = self.sc_sound_queue.qsize()
                    space_left = self.sound_size - qsize
                    if space_left < 2*buffer:
                        cmd = self.sc_sound_queue.get()
                        SlTrace.lg(f"dropping {cmd}")
                    else:
                        break
            
            self.sound_busy = True  # Busy till complete
                                    # Avoid hazard of empty queue
                                    # and no active sound        
            cmd = self.sc_sound_queue.get()
            SlTrace.lg(f"speech queue: cmd: {cmd}", "sound_queue")
            if cmd.cmd_type == "CLEAR":
                continue
            elif cmd.cmd_type == "QUIT":
                self.pyttsx_proc.quit()
                break

            elif cmd.cmd_type == "CMD_MSG":
                msg = cmd.msg
                msg_type = cmd.msg_type
                self.speak_text(msg, msg_type=msg_type,
                                rate=cmd.rate, volume=cmd.volume)
                SlTrace.lg(f"sound_queue: cmd: {cmd} AFTER speak_text", "sound_queue")
            elif cmd.cmd_type == "CMD_TONE":
                self.play_tone(cmd.tone)
                SlTrace.lg(f"sound_queue: cmd: {cmd} AFTER play_tone", "sound_queue")
            elif cmd.cmd_type == "CMD_WAVEFORM":
                self.play_waveform(cmd.waveform)
                SlTrace.lg(f"sound_queue: cmd: {cmd} AFTER play_waveform", "sound_queue")
            else:
                raise SpeakerControlError(f"Unrecognized speaker cmd type:{cmd.cmd_type} in {cmd}")
            if cmd.after is not None:
                cmd_id = cmd.cmd_id
                if cmd.fr is None:
                    ###wxportraise SpeakerControlError(f"after missing fr in cmd {cmd}")
                    pass
                if cmd_id in self.cmds_in_progress:
                    del self.cmds_in_progress[cmd_id]
                
        SlTrace.lg("sc_sound_proc_thread returning")
            
    def quit(self):
        SlTrace.lg("SpeakerControl quitting")
        self.clear()
        self._running = False       # All our threads watch this
        self.sc_cmd_queue.put(SpeakerControlCmd(cmd_type="QUIT"))   # drop the wait
        self.sc_sound_queue.put(SpeakerControlCmd(cmd_type="QUIT"))   # drop the wait
        SlTrace.lg("Force threads stop")
        #self.sc_sound_thread.join()
        #self.sc_cmd_thread.join()

    def delay_for(self, dur):
        """ Wait for dur seconds
        :dur: duration seconds
        """
        self.delay_start(dur=dur)
        self.delay_wait()
        
    def delay_start(self, dur):
        """ Start delay
        :dur: duration in seconds
        """
        self._delay = SpeakerControlDelay(self, dur=dur)

    def delay_wait(self):
        """ wait till delay end
        """
        while not self._delay.is_end():
            time.sleep(.001) 
        
    def play_tone(self, tone):
        """ Called to play pending tone
        :tone: (SpeakerTone) tuple left,right or monoral
        """
        if type(tone.volume) != tuple:
            tone.volume = (tone.volume,tone.volume)
            
        self.sc_tone_busy = True
        SlTrace.lg(f"play_tone: qsize: {self.get_sound_queue_size()}",
                    "sound_queue")
        try:                
            with self.sound_lock:
                vol_left,vol_right = tone.volume
                self.delay_start(tone.delay)
                stereo_waveform = SineWaveNumPy(pitch = tone.pitch,
                    duration=tone.dur,
                    decibels_left=vol_left,
                    decibels_right=vol_right)
                self.delay_wait()
            stereo_waveform.play()
            self.delay_for(dur=tone.dur)
            stereo_waveform.stop()
            
        except Exception as e:
            SlTrace.lg("Bust out of play_tone")
            SlTrace.lg(f"Unexpected exception: {e}")
            SlTrace.lg("Printing the full traceback as if we had not caught it here...")
            SlTrace.lg(format_exception(e))
        self.sc_tone_busy = False
        
    def play_waveform(self, waveform):
        """ Called to play pending waveform
        :waveform: (SpeakerWaveform) waveform to play
        :dur: duration default: waveform.dur
        :sample_rate: sample rate default: waveform.sample_rate,
                                           else: self.sample_rate
        """
        self.sc_tone_busy = True
        dur = waveform.dur
        sample_rate = waveform.sample_rate
        if sample_rate is None:
            sample_rate = self.sample_rate
        if dur is None or waveform.calculate_dur:
            wf_shape = waveform.ndarr.shape
            wf_len = wf_shape[0]
            dur = wf_len/sample_rate
        SlTrace.lg(f"play_waveform: len:{wf_len} dur: {dur}",
                    "sound_time")
        try:                
            with self.sound_lock:
                if waveform.delay is not None:
                    self.delay_for(waveform.delay)
                ts = time.time()
                sd.play(waveform.ndarr, sample_rate)
                self.delay_for(dur=dur)
                sd.stop()
                te = time.time()
                SlTrace.lg(f"play_waveform end tw=({te-ts:.3f})", "sound_time")
        except Exception as e:
            SlTrace.lg("Bust out of play_waveform")
            SlTrace.lg(f"Unexpected exception: {e}")
            SlTrace.lg("Printing the full traceback as if we had not caught it here...")
            SlTrace.lg(format_exception(e))
        self.sc_tone_busy = False
                
    def speak_text(self, msg, msg_type=None,
                   rate=240, volume=.9):
        """ Called to speak pending line
        :msg: text of message
        :mst_type: default: REPORT
        :rate: speech rate words per minute
                default: 240
        :volume: volume default: .9
        
        """
        SlTrace.lg(f"sc: speek_text: qsize: {self.get_sound_queue_size()}",
                    "speech")
        
        if msg_type is None:
            msg_type = "REPORT"
        SlTrace.lg(f"""sc: speak_text(msg={msg}, msg_type={msg_type},"""
                   f""" rate={rate}, volume={volume})""", "speak_text")
        cmd = SpeechMakerCmd(msg=msg, msg_type=msg_type,
                 rate=rate, volume=volume)
        self.pyttsx_proc.make_cmd(cmd)

    def send_cmd(self, cmd_type='speak', msg=None, msg_type=None,
                 rate=None, volume=None, tone=None,
                 waveform=None, fr=None, after=None):
        """ Send cmd to speaker control engine
            storing hash of cmds in process
            Sets self.sc_busy True, which is cleared
            when speech and tone are complete/idle
            and all queues are empty
        :cmd_type: command type
                    'text'        - speak text
                    'tone'        - tone
                    'erase_queue' - erase/cancel pending speech/tones
                    'exit' - erase/cancel pending speech/tones and then stop engine
        :msg: message text, if any
        :msg_type: type of text
                'report'  - standard report
                'command' - a command
        :rate: speaking rate for speech
        :volume: speaking volume for speech
        :tone: SpeakerTone
        :waveform: waveform (SinewaveNumPy.stereo_waveform) NumPy array
        :fr: SpeakerControlLocal's window reference
        :after: function to call after cmd completion
                default: no call
        :returns: cmd, cmd_id, after is needed but cmd is for documentation
        """
        self.sc_busy = True
        SlTrace.lg(f"send_cmd:{cmd_type} msg: {msg} tone: {tone}", "sound_queue")
        cmd = SpeakerControlCmd(cmd_type=cmd_type, msg=msg,
                                 msg_type=msg_type,
                                 rate=rate, volume=volume,
                                 tone=tone, waveform=waveform,
                                 after=after)
        cmd_id = cmd.cmd_id
        self.cmds_in_progress[cmd_id] = cmd
        self.sc_cmd_queue.put(cmd)
        return cmd

class SimpleSpeakerControl():
    """ Simple speaker control for distributed operation
    Mostly for debugging/testing
    """

    def __init__(self):
        self.simple_speaker = True

    def get_cmd_queue_size(self):
        return 0
        
    def is_busy(self):
        return False
    
    def quit(self):
        return
    
    def clear(self):
        return
    
    def send_cmd(self, cmd_type='speak', msg=None, msg_type=None,
                 rate=None, volume=None, tone=None,
                 waveform=None, fr=None, after=None):
        """ Stub out major cmd processing
        """

    def get_sound_queue_size(self):
        return 0
        
    def get_vol_adj(self):
        return 0
    
    def busy_parts(self):
        return False

    def get_sound_queue_size(self):
        """ Get current number of entries
        :returns: number of entries
        """
        return self.sc.get_sound_queue_size()
    
    

            
            
class SpeakerControlLocal:
    """ Localinstance of SpeakerControl
    """

    def __init__(self, logging_sound=False, simple_speaker=False):
        self.make_silent(False)
        self.logging_sound = logging_sound
        self.simple_speaker = simple_speaker
        if simple_speaker:
            self.sc = SimpleSpeakerControl()
            return
        
        self.cmds_awaiting_after = {} # dictionary by cmd_id awaiting after
        self.awaiting_loop_ms = 1      # Awaiting loop
        self.awaiting_loop_going = False    # Checking loop in progress
        self.sc = SpeakerControl()      # NOTE: no simple_speaker arg

    def get_sound_queue_max(self):
        """ Get maximum number of entries
        :returns: max number of entries
        """
        return self.sc.get_sound_queue_max()

    def get_sound_queue_size(self):
        """ Get current number of entries
        :returns: number of entries
        """
        return self.sc.get_sound_queue_size()

    def make_silent(self, val=True):
        self._silent = val

    def clear(self):
        """ Clear pending output
        """
        self.sc.clear()
        self.clear_awaiting()

    def force_clear(self, restart=False):
        """ Clear pending output
        """
        SlTrace.lg("force_clear")
        self.sc.force_clear()
        rwait = 2000
        if restart:
            SlTrace.lg(f"Waiting {rwait} msec")
            wx.CallLater(rwait, self.sc.start_control)

    def get_cmd_queue_size(self):
        """ Get current number of entries
        :returns: number of entries
        """
        return self.sc.get_cmd_queue_size()

    def get_vol_adj(self):
        """ Get current volume adjustment ??? Thread Safe ???
        :returns: current vol_adjustment in db
        """
        return self.sc.get_vol_adj()

    def is_busy(self):
        """ Check if if busy or anything is pending
        """
        return self.sc.is_busy()

    def wait_while_busy(self):
        while self.sc.busy_parts():
            pass    ###wxport###
            
    def quit(self):    
        self.sc.quit()

    def play_tone(self, pitch, dur=None, volume=None, dly=None):
        """ play tone
        :pitch: tone pitch
        :dur: tone duration
        :volume: (left,right) decibels
        :dly: delay(seconds) from start default: no delay
        """
        tone = SpeakerTone(pitch=pitch, dur=dur, volume=volume)
        SlTrace.lg(f"play_tone tone: {tone}", "play_tone")
        self.sc.send_cmd(cmd_type="CMD_TONE", tone=tone)
        
    def play_waveform(self, ndarr, dur=None, dly=None,
                               sample_rate=None,
                               calculate_dur=False,
                               after=None):
        """ play waveform
        :ndarr: waveform (SinewaveNumPy stereo_waveform)
        :dur: wave max duration (seconds)
        :dly: delay(seconds) from start default: no delay
        :sample_rate: sample rate fps
        :calculate_dur:    # calculate duration based on waveform, sample_rate
        :after: function to call after play completes
                 (self.wx_win.after(0,after)
                default: no call
        """
        waveform = SpeakerWaveform(ndarr=ndarr, dur=dur, delay=dly,
                                   calculate_dur=calculate_dur,
                                   sample_rate=sample_rate)
        SlTrace.lg(f"play_waveform waveform: {waveform}", "sound_queue")
        cmd = self.sc.send_cmd(cmd_type="CMD_WAVEFORM", waveform=waveform,
                         after=after)
        if after is not None:
            self.add_awaiting(cmd)

    def add_awaiting(self, cmd):
        """ Add cmds awaiting after calls
            start waiting loop if necessary
        :cmd: cmd sent
        """
        self.cmds_awaiting_after[cmd.cmd_id] = cmd
        if not self.awaiting_loop_going:
            self.awaiting_loop_going = True
            wx.CallAfter(self.awaiting_after_ck)

    def awaiting_after_ck(self):
        """ Check cmds awaiting for after cking
        """
        self.awaiting_loop_going = False    # Set True if more needed
                                            # Delete in order
        cmd_ids  = sorted(list(self.cmds_awaiting_after))   # So we can delete in loop
        for cmd_id in cmd_ids:
            if cmd_id in self.cmds_awaiting_after:
                cmd = self.cmds_awaiting_after[cmd_id]
                if not self.sc.is_in_progress(cmd_id):
                    SlTrace.lg(f"{cmd_id}: after_called for {cmd}", "sound_queue")
                    wx.CallAfter(cmd.after)
                    del self.cmds_awaiting_after[cmd_id]
        if len(self.cmds_awaiting_after) > 0:
            self.awaiting_loop_going = True
            wx.CallLater(self.awaiting_loop_ms, self.awaiting_after_ck)

    def clear_awaiting(self):
        """ Clear out awaiting, calling all awaiting
        """
        self.awaiting_loop_going = False
        if self.simple_speaker:
            return
        
        cmd_ids  = sorted(list(self.cmds_awaiting_after))   # So we can delete in loop
        for cmd_id in cmd_ids:
            if cmd_id in self.cmds_awaiting_after:
                cmd = self.cmds_awaiting_after[cmd_id]
                SlTrace.lg(f"{cmd_id}: after_called for {cmd}", "sound_queue")
                wx.CallAfter(cmd.after)
                del self.cmds_awaiting_after[cmd_id]
                        
    def speak_text(self, msg, dup_stdout=True,
                   msg_type=None,
                   rate=None, volume=None):
        """ Speak text, if possible else write to stdout
        :msg: text message, iff speech
        :dup_stdout: duplicate to stdout default: True
        :msg_type: type of speech default: 'REPORT'
            REPORT - standard reporting
            CMD    - command
            ECHO - echo user input
        :rate: speech rate words per minute
                default: 240
        :volume: volume default: .9
            
        """
        if msg_type is None:
            msg_type = "REPORT"
        if self.logging_sound:
            SlTrace.lg(msg)
        text_lines = msg.split("\n")
        for text_line in text_lines:
            if not self._silent:
                self.sc.send_cmd(cmd_type="CMD_MSG", msg=text_line,
                                  msg_type=msg_type,
                                  rate=rate, volume=volume)
        if dup_stdout and not self.logging_sound:
            SlTrace.lg(msg)

    def stop_speak_text(self):
        """ Stop pending speech
        """
        self.sc.send_cmd(cmd_type="CMD_STOP_SPEAK_TEXT")

    def stop_scan(self):
        """ Stop current scan
        """
        self.sc.stop_scan()
        self.stop_speak_text()    

if __name__ == "__main__":
    import os
    import time
    import multiprocessing
    multiprocessing.freeze_support()
    from wx_win import WxWin
    adw = None
    wx_win = WxWin(adw, "wx_speaker_control Self Test")
    for simple_speaker in [True, False]:
        SlTrace.lg(f"simple_speaker: {simple_speaker}")
        SlTrace.setFlags("speech,sound_queue")
        scl = SpeakerControlLocal(simple_speaker=simple_speaker)
        scl.wait_while_busy()
        long_msg = """
        line one
        line two
        line three
        line four
        line five
        line six
        line seven
        line eight
        line nine
        line ten
        """
        long_msg_group = long_msg.split("\n")
        for msg in ["one"]:
            scl.speak_text(msg)
            time.sleep(1)
            scl.wait_while_busy()

            scl.speak_text(long_msg)
            scl.wait_while_busy()
            
            scl.speak_text("Hello World!")
            scl.wait_while_busy()
            scl.speak_text("How are you?")
            scl.wait_while_busy()
            scl.speak_text("Hows the weather?")
            scl.wait_while_busy()
            scl.speak_text("What's up?")
            time.sleep(3)
            scl.clear()
            scl.speak_text("Just cleared")
            time.sleep(2)
            scl.wait_while_busy()



#pyttsx_proc.py   19Sep2023  crs
""" pyttsxN ecapuslated in a separate process
    Supports
        1. queued requests
        2. clearing pending and current requests
"""
import multiprocessing as mp
import time

from speech_maker_cmd import SpeechMakerCmd
#import pyttsx3 as pyttsxN    # Can use pyttsx3 instead
import pyttsx4 as pyttsxN    # Can use pyttsx4 instead

from select_trace import SlTrace

class PyttsxProc:
    def __init__(self, qlen=100):    # 10 USUALY
        """ Setup for interprocess communication
        :qlen: input/output queue length
                default: 10
        """
        self.qlen = qlen
        self.setup_proc()       # setup incase
                                # is_busy called
                                # before talk
    def pyt_proc_proc(self):
        """ process speech commands
            process's processing procedure
        """    
        SlTrace.clearFlags()    # Clear flags in separate process
        self.engine = pyttsxN.init()

        while True:
            cmd = self.pyt_queue.get()
            self.do_cmd(cmd)

    def setup_proc(self):
        """ Setup procesing process
        """
        self.pyt_proc = mp.Process(target=self.pyt_proc_proc)
        self.pyt_queue = mp.Queue(self.qlen)  # speech queue of SpeakerControlCmd
        self.pyt_out_queue = mp.Queue(2*self.qlen)
        self.pyt_engine_busy = False   # Updated with status
        time.sleep(1) 
        self.pyt_proc.start()
        
    def clear(self):
        """ Clear current and pending talking
        """
        if self.pyt_proc is None:
            return      # clear or unset
        
        self.pyt_proc.kill()
        self.pyt_proc.join()
        self.pyt_proc = None    # Forces talk to re-setup
        
                    
    def is_alive(self):
        """ Check if speech processing is going
        """
        if self.pyt_proc is None:
            return False    # inactive
        
        return self.pyt_proc.is_alive()

    
    def talk(self, text, rate=None, volume=None):
        """ Say text
        Starts process if inactive
        :text: text to speak
        """
        cmd = SpeechMakerCmd(msg=text, rate=rate,
                             volume=volume)
        self.make_cmd(cmd)
    
    def make_cmd(self, cmd):
        """Make speech command
        :cmd: SpeechMakerCmd
            QUIT - clear pending and current speech
        """
        if self.pyt_proc is None:
            self.setup_proc()
            
        if cmd.cmd_type == "QUIT":
            self.quit()
            return
        
        while self.pyt_queue.qsize() >= self.qlen:
            SlTrace.lg(f"talk queue size "
                       f" {self.pyt_queue.qsize()}"
                       f" dropping {cmd.msg}")
            return
        self.pyt_queue.put(cmd)
        SlTrace.lg(f"make_cmd - queue size: {self.pyt_queue.qsize()}", "talk_cmd")
            
    def do_cmd(self, cmd):
        """ do talking, waiting till done
        :cmd: SpeechMakerCmd
        """
        self.engine_runAndWait(cmd)
        SlTrace.lg("after engine_runAndWait()", "talk_cmd")
        SlTrace.lg(f"do_cmd - queue size: {self.pyt_queue.qsize()}", "talk_cmd")

    def engine_runAndWait(self, cmd):
        """ run engine, wait till done
        """
        self.pyt_out_queue.put(True)    # Set as busy
        if cmd.rate is not None:
            self.engine.setProperty('rate', cmd.rate)
        if cmd.volume is not None:
            self.engine.setProperty('volume', cmd.volume)
        if cmd.msg is not None and cmd.msg != '':
            self.engine.say(cmd.msg)
            SlTrace.lg(f"After eng.say({cmd.msg})", "talk_cmd")
            self.engine.runAndWait()
        SlTrace.lg("after eng.runAndWait()", "talk_cmd")
        self.pyt_out_queue.put(False)
        
    def wait_while_busy(self):
        while True:
            if not self.is_busy():
                break

    def is_busy(self):
        """ Check if busy talking or
        getting ready to talk
        """
        if not self.is_alive():
            return False
        
        if self.pyt_queue.qsize()>0:
            return True
        
        while self.pyt_out_queue.qsize() > 0:
            if not self.is_alive():
                return False
            self.pyt_engine_busy = self.pyt_out_queue.get()
        return self.pyt_engine_busy
        
    
    def quit(self, wait=True):
        """ Quit talking
        :wait: if true wait till talking is done
                default: True - wait
        """
        SlTrace.lg("Quitting", "talk_cmd")
        if wait:
            self.wait_while_busy()
        self.clear()
        SlTrace.lg("After quit", "talk_cmd")
           
    
if __name__ == "__main__":
    import time
    
    mp.freeze_support()
    SlTrace.clearFlags()
    #SlTrace.setFlags("talk_cmd")
    
    SlTrace.lg("\nTest Start")
    tt = PyttsxProc()
    
    # Test from wx_speaker_control.py
    # Looking at bug: talking stops after 4th line
    #
    scl = PyttsxProc()
    long_msg = """
    line one
    line two
    line three
    line four
    line five
    line six
    line seven
    line eight
    line nine
    line ten
    """
    long_msg_group = long_msg.split("\n")
    for msg in long_msg_group:
        scl.talk(msg)
        time.sleep(1)
    scl.wait_while_busy()   # Replaces wx_win.MainLoop()

    scl.talk(long_msg)
    scl.wait_while_busy()   # Replaces wx_win.MainLoop()
    
    
    tt.talk("Hello World")
    for i in range(8):
        print(i, "\a")
        time.sleep(1)
    tt.talk(" ".join([str(i) for i in range(100,0,-1)]))
    time.sleep(4)
    tt.clear()
    tt.talk("After clear")
    tt.talk("How are you?", volume=.5)
    tt.talk("How's the weather?", rate=120)
    tt.talk("Good Bye for now.", volume=.99)
    tt.wait_while_busy()
    tt.quit()
    SlTrace.lg("Test End\n")


# wx_canvas_panel.py 28Nov2023  crs, rename canvas_panel.py
#                    08Nov2023  crs
"""
Require AudioDrawWindow as parent
We may make this a derived class AdwCanvasPanel of CanvasPanel
in the future
Support for very limited list of tkinter Canvas type actions
on wxPython Panel  Our attempt here was to ease a port
from tkinter Canvas use to wxPython.
"""
import time
import wx
from select_trace import SlTrace, SelectError
from wx_stuff import * 
from wx_canvas_panel_item import CanvasPanelItem            
from wx_adw_display_pending import AdwDisplayPending

        
class CanvasPanel(wx.Panel):
    
    """ Panel in which we can do tkinter like things
    """
    def __init__(self, frame, app=None, color=None,
                key_press_proc=None,
                **kw):
        """
        :frame: parent frame
        :app: wx.App, default create
        :color: background color default: :" light gray"
        :key_press: function(sym) to execute
                    keyboard command
                    default: echo sym
        """
        self.adw = self.frame = frame
        self._cursor_rect = None    # for erasing old cursor
        
        self.sfx =  1       # default scaling
        self.sfy = 1
        if app is None:
            app = wx.App()
        self.app = app
        super().__init__(frame, **kw)
        self.time_of_update = time.time()
        self.min_time_update = .01      # min time between updates (seconds)
        self.mouse_left_down_proc = self.mouse_left_down_def
        self.mouse_motion_proc = self.mouse_motion_def
        self.mouse_b1_motion_proc = self.mouse_b1_motion_def
        if key_press_proc is None:
            key_press_proc = self.key_press
        self.key_press_proc = key_press_proc
        if color is None:
            color = "light gray"
        self.SetBackgroundColour(color)
        self.color = color
        self.frame_size = frame.GetSize()
        self.text_entry_panel = wx.Panel(self)
        self.grid_panel = wx.Panel(self)
        self.entry = wx.TextCtrl(self.text_entry_panel,
                                 size=wx.Size(300,20),
                                 style=wx.TE_PROCESS_ENTER)
        sizer_v = wx.BoxSizer(wx.VERTICAL)
        sizer_v.Add(self.text_entry_panel, 0, wx.ALIGN_CENTER_HORIZONTAL)
        sizer_v.Add(self.grid_panel, 1, wx.EXPAND)
        self.SetSizer(sizer_v)
        self.grid_panel_offset = 0 # menu + text TBD
        self.Show()
        self.can_id = 0
        self.items_by_id = {}   # Items stored by item.canv_id
                                # Augmented by CanvasPanelItem.__init__()
        self.items = []         # Items in order drawn
        self.adw_dp = AdwDisplayPending(self)        
        self.prev_reg = None    # Previously displayed

        #self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.grid_panel.Bind(wx.EVT_PAINT, self.OnPaint)
        self.grid_panel.Bind(wx.EVT_SIZE, self.OnSize)
        self.grid_panel.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
        self.grid_panel.Bind(wx.EVT_CHAR_HOOK, self.on_char_hook)
        
        self.motion_level = 0   # Track possible recursive calls
        self.grid_panel.Bind(wx.EVT_MOTION, self.on_mouse_motion)
        self.frame.Bind(wx.EVT_LEFT_DOWN, self.on_mouse_left_down_frame)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_mouse_left_down_win)
        self.grid_panel.Bind(wx.EVT_LEFT_DOWN, self.on_mouse_left_down)
        self._multi_key_progress = False    # True - processing multiple keys
        self._multi_key_cmd = None          # Set if in progress
        
        
        self.Show()
        self.on_paint_skip = 1 # debug - update only every nth
        self.on_paint_count = 0 # count paints
        frame.SetSize(self.frame_size)
        panel_size = frame.GetClientSize()
        frame.Show()
        self.SetSize(panel_size)
        self.Refresh()
        self.Update()
        self.Show()
        wx.CallLater(0, self.SetSize, (self.frame.GetSize()))
        self.Show()
        self.grid_panel.SetFocus() # Give grid_panel focus
        
    def scale_points(self, pts, sfx=None, sfy=None):
        """ scale a list of two dimensional points
        :pts: list of wx.Points
        :sfx: x scale factor  default: 1
        :sfy: y scale factor default: 1
        :returns: scaled points list
        """
        """ TFD Return original points unscalled
        """
        use_orig_points = True
        use_orig_points = False
        if use_orig_points:
            SlTrace.lg("use_orig_points", "paint")
            return pts[:]   # Copy of original points, unchanged
        
        if sfx is None:
            sfx = self.sfx
        if sfy is None:
            sfy = self.sfy
        pts_s = []
        if len(pts) == 0:
            return pts_s
        
        scale_1to1 = True
        if scale_1to1:
            sfx = sfy = 1.0
            SlTrace.lg(f"Force sfx:{sfx} sfy:{sfy}", "paint")
        x0,y0 = pts[0]
        for i in range(len(pts)):
            x,y = pts[i]
            x_s = x0 + int((x-x0)*sfx)
            y_s = y0 + int((y-y0)*sfy)
            pts_s.append(wx_Point(x_s,y_s))
        return pts_s
      
    def get_id(self):
        """ Get unique item id id
        """
        self.can_id += 1
        return self.can_id

    def no_call(self):
        """ Just a nothing call
        """
        return
    
    def after(self, delay, callback=None):
        """ Call after delay
        :delay: delay in milliseconds
        :callback: function to call
            default: no function, just delay
        """
        if callback is None:
            callback = self.no_call
        wx.CallLater(delay, callback)

    def create_composite(self, disp_type=None, desc=None):
        """ Create a composite display object
        whose components are displayed in order as an object
        :desc: description, if one
        """
        item = CanvasPanelItem(self, "create_composite",
                               disp_type=disp_type, desc=desc)
        self.items.append(item)
        self.items_by_id[item.canv_id] = item    # store by id
        return item.canv_id
                    
    def create_rectangle(self, cx1,cy1,cx2,cy2,
                                **kwargs):
        """ Implement create_rectangle
            supporting: fill, outline, width
        :returns: id
        """
        item = CanvasPanelItem(self, "create_rectangle",
                               cx1,cy1,cx2,cy2,
                               **kwargs)
        self.items.append(item)
        self.items_by_id[item.canv_id] = item    # store by id
        return item.canv_id
                    
    def create_bitmap(self, cx1,cy1,cx2,cy2,
                      bitmap, desc=None,
                                **kwargs):
        """ Implement create_bitmap
        :returns: id
        """
        item = CanvasPanelItem(self, "create_bitmap",
                               cx1,cy1,cx2,cy2,
                               bitmap=bitmap,
                               desc=desc,
                               **kwargs)
        self.items.append(item)
        self.items_by_id[item.canv_id] = item    # store by id
        return item.canv_id
    
    def create_cursor(self, x0,y0,x1,y1,
                        **kwargs):
        """ Implement create_cursor
            supporting fill, outline, width
        """
        item = CanvasPanelItem(self, "create_cursor",
                               x0,y0,x1,y1,
                               **kwargs)
        self.items_by_id[item.canv_id] = item
        self.items.append(item)
        return item.canv_id
    
    def create_oval(self, x0,y0,x1,y1,
                        **kwargs):
        """ Implement create_oval
            supporting fill, outline, width
        """
        item = CanvasPanelItem(self, "create_oval",
                               x0,y0,x1,y1,
                               **kwargs)
        self.items_by_id[item.canv_id] = item
        self.items.append(item)
        return item.canv_id
    
    def create_point(self, x0,y0, radius=2,
                        **kwargs):
        """ Helper function to create a point
            Shortcut using create_oval
            supporting fill, outline, width
        """
        px0 = x0-radius
        py0 = y0-radius
        px1 = x0+radius
        py1 = y0+radius
        item_id = self.create_oval(px0,py0,px1,py1,
                                **kwargs)
        return item_id

    def create_line(self, *args, **kwargs):
        """ Implement tkinter's create_line
        :args: x1,y1,...xn,yn
        :kwargs:  width=, fill=, tags=[]
        """
        item = CanvasPanelItem(self, "create_line",
                               *args,
                               **kwargs)
        self.items_by_id[item.canv_id] = item
        self.items.append(item)
        return item.canv_id
    
    def create_text(self, x0,y0,
                        **kwargs):
        """ Implement create_text
            supporting fill, font
        """
        item = CanvasPanelItem(self, "create_text",
                               x0,y0,
                               **kwargs)
        self.items_by_id[item.canv_id] = item
        self.items.append(item)
        return item.canv_id
        

    def delete(self, *id_tags):
        """ Delete object(s) in panel
        :id_tags: if str: "all" - all items, else tag
                else id
        """
        for id_tag in id_tags:
            self.delete_id(id_tag)

    def delete_id(self, id_tag):
        """ Delete items having id or tag
        :id_tag: id or tag
        """
        for item in self.items:
            if type(id_tag) == int:
                if id_tag == item.canv_id:
                    item.deleted = True
                    item.refresh()          # Force redraw
                    break   # id is unique
            
                if id_tag in item.tags:
                    item.deleted = True

    def OnSize(self, e):
        self.Refresh()
        self.Update()
        SlTrace.lg(f"\nOnSize paint count:{self.on_paint_count}", "paint")
        size = self.GetSize()
        SlTrace.lg(f"panel size: {size}", "paint")
        e.Skip()

    def draw_item(self, item):
        """ Draw item
        :item: CanvasPanalItem/canv_id item to draw
        """
        if type(item) == int:
            item = self.id_to_item(item)
        item.draw()
        
    def draw_items(self, items=None, rect=None,
                   types="create_composite"):
        """ Draw scalled items
        :items: items to draw default: self.items
        :rect: Draw only items in this rectangle
                default: draw all items
        :types: item types to draw
                "ALL" - all types except "create_composite"
                default: "create_composite"
        """
        if items is None:
            items = self.items
        if len(items) == 0:
            return
        do_composite = False
        do_all = False
        if types == "create_composite":
            do_composite = True
        elif types == "ALL":
            do_all = True
        else:
            if not isinstance(types,list):
                types = [types]     # Make list
        self.pos_adj = self.cur_pos-self.orig_pos
        SlTrace.lg(f"orig_size: {self.orig_size}"
                   f" cur_size: {self.cur_size}", "item")
        self.sfx = self.cur_size.x/self.orig_size.x
        self.sfy = self.cur_size.y/self.orig_size.y
        if len(items) == 0:
            return      # Short circuit if no items
        
        items_points = self.get_items_points(items)
        items_points_scaled = self.scale_points(items_points)

        ipo = 0     # current offset into items_points_scaled
        for item in items:                
            npts = len(item.points)
            points = items_points_scaled[ipo:ipo+npts]
            SlTrace.lg(f"item: {item}", "item")
            if ((do_composite and item.canv_type == "create_composite")
                    or (do_all and item.canv_type != "create_composite")
                    or (item.canv_type in types)):
                item.draw(points=points, rect=rect)
            ipo += npts # move to next item

    def get_items_points(self, items=None):
        """ Get all drawing points, or embeded figures
        characterization
        points
        :items: items list default: self.items
        :returns: list of all drawn points
        """
        if items is None:
            items = self.items
        points = []
        for item in self.items:
            points += item.points
        return points
                    
    def OnPaint(self, e):
        """ Response to paint event
        """
        dc = wx.PaintDC(self.grid_panel)
        self.on_paint_count += 1
        SlTrace.lg(f"\nOnPaint {self.on_paint_count}", "paint")
        size = self.GetSize()
        SlTrace.lg(f"panel size: {size}", "paint")
        
            
        if self.on_paint_count == 1:
            self.orig_pos = self.GetPosition()  # After setup
            self.orig_size = self.GetSize()
            SlTrace.lg(f"Set orig_size:{self.orig_size}", "paint")
            self.cur_pos = self.prev_pos = self.orig_pos
            self.cur_size = self.prev_size = self.orig_size
            if self.orig_size[0] < 50:
                self.orig_size = self.frame.GetClientSize()
                SlTrace.lg(f"Set orig_size:{self.orig_size}", "paint")
                SlTrace.lg(f"use client size: {self.orig_size}", "paint")
                self.cur_size = self.orig_size
                SlTrace.lg(f"Set cur size: {self.cur_size}", "paint")

            SlTrace.lg(f"First Paint size: {self.orig_size}", "paint")
            SlTrace.lg(f"Frame size: {self.frame.GetSize()}", "paint")
            SlTrace.lg(f"Frame size: {self.frame.GetSize()}", "paint")
            self.Show()
            SlTrace.lg(f"First Paint size: {self.GetSize()}", "paint")
            SlTrace.lg(f"Frame size: {self.frame.GetSize()}", "paint")
            return
                
        else:
            self.cur_size = self.GetSize()
            if self.cur_size[0] < 50:
                self.prev_size = self.cur_size = self.frame.GetClientSize()
                SlTrace.lg(f"use client size: {self.cur_size}", "paint")
                SlTrace.lg(f"Set cur size: {self.cur_size}", "paint")
            SlTrace.lg(f"Further Paint size: {self.GetSize()}", "paint")
            SlTrace.lg(f"Frame size: {self.frame.GetSize()}", "paint")
            pass

        self.display_pending(dc)
#        self.check_for_display()    # TFD - wait while events are processed
        self.prev_pos = self.cur_pos
        self.prev_size = self.cur_size
        SlTrace.lg(f"OnPaint: {self.cur_pos} {self.cur_size}", "paint")
        self.display_cursor(dc)     # Display cursor if any

    def display_cursor(self, dc):
        """ Display cursor, if any
        This is called at end of OnPaint 
        :dc: current display context (wx.PaintDC)
        """
        adw = self.adw
        if adw._cursor_xy is None:
            return
        
        pos_x,pos_y = adw._cursor_xy
        x = int(pos_x)
        y = int(pos_y)
        outline = "black"
        fill = "red"
        dc.SetPen(wx.Pen(outline, style=wx.SOLID))
        dc.SetBrush(wx.Brush(fill, wx.SOLID))
        radius = 5
        pt = wx.Point(x,y)
        dc.DrawCircle(pt=pt, radius=radius)
        sur = 10*radius  # extra
        self._cursor_rect = wx.Rect(
                wx.Point(x-sur, y-sur),
                wx.Point(x+sur, y+sur))
                                    
    def refresh_cursor(self):
        """ Refresh previous cursor if any
        """
        if self._cursor_rect is not None:
            self.RefreshRect(self._cursor_rect)
            #self.Refresh()
            self._cursor_rect = None
            
        
    def set_check_proceed(self, proceed=True):
        self.check_proceed = proceed

    def my_app_kill(self, my_app):
        """ stop sup application loop
        :my_app: local wx.App instance
        """         
        my_app.ExitMainLoop()
        
    def check_for_display(self):
        """ Debug wait with event processing
        """
        self.app.MainLoop()

        duration = 4
        self.check_proceed = False
        wx.CallLater(duration*1000, self.set_check_proceed)
        while not self.check_proceed:
            my_app = wx.App()
            wx.CallLater(10, self.my_app_kill, my_app=my_app)
            my_app.MainLoop()
            pass

    def clear(self, refresh=True):
        """ Clear panel
        :refresh: True - refresh after clear
        """
        self.items_by_id = {}   # Items stored by item.canv_id
                                # Augmented by CanvasPanelItem.__init__()
        self.items = []         # Items in order drawn
        self.prev_reg = None    # Previously displayed
        self.adw_dp.clear()
        #self.Refresh()
        
                
                    
    def itemconfig(self, tags, **kwargs):
        """ Adjust items with tags with kwargs
        :tags: tag or tag list of items to change
        :kwargs: new attributes
                    supporting: outline
        """
        if type(tags) != list:
            tags = [tags]
            
        for item in self.items:
            ins = item.tags.intersection(tags)
            if len(ins) > 0:   # item have tag?
                for kw in  kwargs:
                    val = kwargs[kw]
                    if kw == "outline":
                        item.kwargs[kw] = val
                    else:
                        raise SelectError(f"itemconfig doesn't support {kw} (val:{val})")    

    
    def update_item(self, item, **kwargs):
        """ Update cell, with "inplace" attribute changes
            refreshes item's region 
        :item: item / id
        :**kwargs: attributes to be changed
        """
        if type(item) == int:
            item = self.items_by_id[item]
        item.update(**kwargs)
        bdrect = item.bounding_rect()
        self.grid_panel.RefreshRect(rect=bdrect)


    """
    ----------------------- Mouse Control --------------------
    """
    def get_panel_loc(self, e):
        """ Get mouse location in panel coordinates
        """
        screen_loc = e.Position
        panel_loc = wx_Point(screen_loc.x,
                            screen_loc.y-
                            self.grid_panel_offset)
        return panel_loc

    def id_to_item(self, canv_id):
        """ Return item, given id
        :returns: canv item
        """
        return self.items_by_id[canv_id]
    
    def get_panel_items(self, loc, canv_type="create_rectangle"):
        """ Get item/items at this location
        :loc: wx.Point location at this point
        :canv_type: canvas type e.g., create_rectangle
                default: create_rectangle
        """
        items = []
        for item in self.items:
            if item.bounding_rect().Contains(loc):
                if item.canv_type == canv_type:
                    items.append(item)
        return items
        
    def get_screen_loc(self, panel_loc):
        """ Convert panel location to screen location
        :panel_loc: wx_Point of location
        :returns: wx_Point on screen
        """
        screen_loc = wx_Point(panel_loc.x,
                              panel_loc.y+
                              self.grid_panel_offset)
        return screen_loc
    
    #import pyautogui   #??? comment this line
                       #??? and first click causes
                       #??? the window to shrink
    # mouse_left_down
    def on_mouse_left_down(self, e):
        """ Mouse down
        """
                
        loc = self.get_panel_loc(e) # grid relative
        SlTrace.lg(f"\non_mouse_left_down panel({loc.x},{loc.y})"
                   f" [{e.Position.x}, {e.Position.y}]"
                   f"window: pos: {self.GetPosition()}"
                   f" size: {self.GetSize()}"
                   f" GetClientAreaOrigin: {self.GetClientAreaOrigin()}", "paint"
                   )
        #e.Skip()
        size = self.grid_panel.GetSize()
        pts = self.scale_points([wx_Point(e.Position.x, e.Position.y),
                                 wx_Point(0,0), wx_Point(size.x,0),
                                 wx_Point(0,size.y), wx_Point(size.x,size.y)])
        pt = pts[0]
        self.mouse_left_down_proc(pt.x, pt.y)
        self.grid_panel.SetFocus() # Give grid_panel focus


    def on_mouse_left_down_win(self, e):
        """ Mouse down in window
        """
                
        loc = wx_Point(e.Position)
        SlTrace.lg(f"\non_mouse_left_down_win ({loc.x},{loc.y})"
                   f" [{e.Position.x}, {e.Position.y}]"
                   f"window: pos: {self.GetPosition()}"
                   f" size: {self.GetSize()}"
                   f" GetClientAreaOrigin: {self.GetClientAreaOrigin()}", "paint"
                   )
        e.Skip()

    def on_mouse_left_down_frame(self, e):
        """ Mouse down in window
        """
                
        loc = wx_Point(e.Position)
        SlTrace.lg(f"\non_mouse_left_down_frame ({loc.x},{loc.y})"
                   f" [{e.Position.x}, {e.Position.y}]"
                   f"window: pos: {self.GetPosition()}"
                   f" size: {self.GetSize()}"
                   f" GetClientAreaOrigin: {self.GetClientAreaOrigin()}", "paint"
                   )
        e.Skip()
        
    def mouse_left_down_def(self, x, y):
        """ process mouse left down event
        Replaced for external processing
        :x: mouse x coordiante
        :y: mouse y coordinate
        """
        SlTrace.lg(f"mouse_left_down_proc x={x}, y={y}", "mouse_proc")
    
    def set_mouse_left_down_proc(self, proc):
        """ Set link to front end processing of
        mouse left down event
        :proc: mouse processing function type proc(x,y)
        """
        self.mouse_left_down_proc = proc

    # mouse motion and mouse down motion
    def on_mouse_motion(self, e):
        """ Mouse motion in  window
        we convert <B1-Motion> into on_button_1_motion calls
        """
        loc = self.get_panel_loc(e)
        if e.Dragging():
            self.mouse_b1_motion_proc(loc.x, loc.y)
        else:
            self.mouse_motion_proc(loc.x, loc.y)

    def mouse_motion_def(self, x, y):
        """ Set to connect to remote processing
        """
        SlTrace.lg(f"mouse_motion_proc(x={x}, y={y})", "motion")

    def mouse_b1_motion_def(self, x, y):
        """ Default mouse_b1_motion event proceessing
        """
        SlTrace.lg(f"mouse_b1_motion_proc(x={x}, y={y})")
    
    def set_mouse_motion_proc(self, proc):
        """ Default mouse_motion event processing
        mouse motion event
        :proc: mouse processing function type proc(x,y)
        """
        self.mouse_motion_proc = proc
   
    def set_mouse_b1_motion_proc(self, proc):
        """ Set link to front end processing of
        mouse b1 down motion event
        :proc: mouse processing function type proc(x,y)
        """
        self.mouse_b1_motion_proc = proc

    def redraw(self):
        """ Redraw screen
        """
        #self.Refresh()
        #self.Update()
        #self.Show()

    def refresh_item(self, item):        
        """ mark item's region to be displayed
        :item: CanvasPanelItem/id
        """
        if type(item) == int:
            if item >= len(self.items):
                return      # Outside bounds - ignore
            
            item = self.items[item]
        rect = item.bounding_rect()
        self.grid_panel.RefreshRect(rect)
                    
    def refresh_rectangle(self, *args,
                                **kwargs):
        """ Mark rectangle in need of repainting
        :arg[0]: wx.Rect 0 rectangle to refresh
        :
        :args[0]..args[3]: rectangle x1,y1, x2,y2 coordinates
        """
        if isinstance(args[0], wx.Rect):
            rect = args[0]
        else:
            rect = wx.Rect(wx.Point(args[0],args[1]),wx.Point(args[2],args[3]))
        SlTrace.lg(f"refresh_rectangle({rect})", "refresh")
        self.grid_panel.RefreshRect(rect)

    def update(self, x1=None, y1=None, x2=None, y2=None,
               full=False):
        """ Update display
            If x1,...y2 are present - limit update to rectangle
            If x1 is a wx.Rect use Rect
            :full: force full update
        """
        return  # TFD
        now = time.time()
        '''
        since_last = now - self.time_of_update
        if since_last < self.min_time_update:
            delay_ms = int((self.time_of_update + self.min_time_update-now)*1000)
            wx.CallLater(delay_ms, self.update, x1=x1, y1=y1, x2=x2, y2=y2)
            return          # Too soon for update
        '''
        self.time_of_update = time.time()
        if full:
            self.grid_panel.Refresh()
            self.grid_panel.Update()
            return
    
        SlTrace.lg(f"update: refresh({x1,y1, x2,y2})", "refresh")
        if x1 is not None or x2 is not None:
            if isinstance(x1, wx.Rect):
                self.grid_panel.RefreshRect(x1)
            else:
                self.grid_panel.RefreshRect(rect=wx.Rect(wx_Point(x1,y1),
                                                         wx_Point(x2,y2)))
        else:
            self.grid_panel.Refresh()
        self.grid_panel.Update()
        
    """
    ----------------------- Keyboard Control --------------------
    """

    def set_key_press_proc(self, proc):
        """ Set key press processing function to facilitating 
        setting/changing processing function to after initial
        setup
        :proc: key processing function type proc(keysym)
        """
        self.key_press_proc = proc

       
    def on_char_hook(self, e):
        SlTrace.lg(f"\non_char_hook:{e}", "keys")
        
        if e.AltDown():
            SlTrace.lg(f"{self.get_mod_str(e)}", "keys")
            e.Skip()    # Pass up to e.g., Menu
            return
        
        SlTrace.lg(f"sym: {self.get_mod_str(e)}", "keys")
        SlTrace.lg(f"GetUnicodeKey:{e.GetUnicodeKey()}", "keys")
        SlTrace.lg(f"GetRawKeyCode:{e.GetRawKeyCode()}", "keys")
        SlTrace.lg(f"GetKeyCode:{e.GetKeyCode()}", "keys")
        SlTrace.lg(f"chr(GetKeyCode){chr(e.GetKeyCode())}"
                   f" {ord(chr(e.GetKeyCode()))}", "keys")
        
        keysym = self.get_keysym(e)
        self.key_press_proc(keysym)
        self.grid_panel.SetFocus() # Give grid_panel focus

    def on_key_down(self, e):
        SlTrace.lg(f"on_key_down:{e}", "keys")
        SlTrace.lg(f"sym: {self.get_mod_str(e)}", "keys")
        if e.AltDown():
            SlTrace.lg(f"{self.get_mod_str(e)}", "keys")
            e.Skip()    # Pass up to e.g., Menu
            return
        
        SlTrace.lg(f"GetUnicodeKey:{e.GetUnicodeKey()}", "keys")
        SlTrace.lg(f"GetRawKeyCode:{e.GetRawKeyCode()}", "keys")
        SlTrace.lg(f"GetKeyCode:{e.GetKeyCode()}", "keys")
        keysym = self.get_keysym(e)
        self.key_press_proc(keysym)

    def get_mod_str(self, e):
        """ return modifier list string
        :e: event
        :returns: "[mod1-][mod2-][...]key_sym"
        """
        mod_str = ""
        if e.HasModifiers():
            mods = e.GetModifiers()
            if mods & wx.MOD_ALT:
                if mod_str != "": mod_str += "-"
                mod_str += "ALT"
            if mods & wx.MOD_CONTROL:
                if mod_str != "": mod_str += "-"
                mod_str += "CTL"
            if mods & wx.MOD_SHIFT:
                if mod_str != "": mod_str += "-"
                mod_str += "SHIFT"
            if mods & wx.MOD_ALTGR and mods == wx.MOD_ALTGR:
                if mod_str != "": mod_str += "-"
                mod_str += "ALTGR"
            if mods & wx.MOD_META:
                if mod_str != "": mod_str += "-"
                mod_str += "META"
            if mods & wx.MOD_WIN:
                if mod_str != "": mod_str += "-"
                mod_str += "WIN"
        ret = mod_str
        if ret != "": ret += "-"
        ret += self.get_keysym(e)
        return ret
        

    def key_press(self, keysym):
        """ default/null simulated key event
        :keysym: Symbolic key value/string
        """
        SlTrace.lg(keysym)
        return

    def get_keysym(self, e):
        """ Convert key symbol to keysym(tkinter style)
        :e: key event
        :returns: keysym(tkinter) string
        """
        unicode = e.GetUnicodeKey()
        raw_key_code = e.GetRawKeyCode()
        key_code = e.GetKeyCode()
        ch = chr(unicode)  # lazy - accept all single char

        if (key_code >= wx.WXK_NUMPAD0 and
            key_code <= wx.WXK_NUMPAD9):
            return str(key_code-wx.WXK_NUMPAD0)        
        if key_code == wx.WXK_ALT:
            return 'Alt_L'
        if key_code == wx.WXK_ESCAPE:
            return 'Escape'
        if key_code == wx.WXK_UP:
            return 'Up'
        if key_code == wx.WXK_DOWN:
            return 'Down'
        if key_code == wx.WXK_LEFT:
            return 'Left'
        if key_code == wx.WXK_RIGHT:
            return 'Right'
        if key_code == wx.WXK_WINDOWS_LEFT:
            return 'win_l'

        if len(ch) == 1:
            return ch
        
        return '???'    # Unrecognized


    """
    Links to adw_display_pending
    """
    
    def add_item(self, item):
        """ Add display item
        :item: item/id (CanvasPanelItem)
        """
        self.adw_dp.add_item(item)

    def get_displayed_items(self):
        """ Get list of displayed items
        :returns: list of permanently displayed values (AdwDisplayPendingItem)
        """
        return self.adw_dp.get_displayed_items()


    def is_overlapping(self, item1, item2):
        """ Check if two display items are overlapping
        :returns: True iff overlapping
        """
        return self.adw_dp.is_overlapping(item1, item2)

        
    def add_cell(self, di_item):
        """ Add cell to be displayed
        :di_item: display item
        """            
        self.adw_dp.add_cell(di_item)

    def add_cursor(self, cursor):
        """ Display list and clear it
        """
        self.adw_dp.add_cursor(cursor)

    
    def display_pending(self, dc):
        """ Display list and clear it
        :dc: wx.PaintDC(self)
        """
        self.adw_dp.display_pending(dc)

if __name__ == "__main__":
    add_menus = True     # True add menus to frame
    add_after_test = True   # Test delay
    add_after_test = False
    
    if add_menus:
        # Provide Menubar to capture Alt-<key>s
        from wx_adw_menus import AdwMenus, FteFake
        
    SlTrace.clearFlags()
    #SlTrace.setFlags("paint")
    SlTrace.setFlags("keys")
    app = wx.App()
    mytitle = "CanvasPanel Selftest"
    width = 600
    height = 800
    if add_menus:
        mytitle += " Using AdwMenus"
        width += 100
    frame = wx.Frame(None, title=mytitle, size=wx.Size(width,height))
    frame.Show()
    if add_menus:
        fte = FteFake()
        menus = AdwMenus(fte, frame=frame)
    canv_pan = CanvasPanel(frame, app=app)
    canv_pan.Show()

    """
    Setup frame level key processing
    Exercising keys passed up from CanvasPanel
    """    
    def frame_on_key_down(e):
        """ process frame level key presses
        """
        SlTrace.lg(f"frame: {canv_pan.get_mod_str(e)}")
        
    def frame_on_char_hook(e):
        """ process frame level key presses
        """
        SlTrace.lg(f"frame hook: {canv_pan.get_mod_str(e)}")
    
    def key_press_proc(key_sym):
        """ Process keys sent from CanvasPanel
        """
        SlTrace.lg(f"key_press: {key_sym}")
        
    if not add_menus:   # if no menu to catch    
        frame.Bind(wx.EVT_KEY_DOWN, frame_on_key_down)
        frame.Bind(wx.EVT_CHAR_HOOK, frame_on_char_hook)

    
    canv_pan.create_rectangle(50,100,200,200, fill="red")
    canv_pan.create_rectangle(150,150,300,300, fill="blue")
    canv_pan.create_oval(50,200,100,300, fill="orange")
    canv_pan.create_oval(150,300,250,350, fill="purple")
    square_size = 100
    for row in range(1,height//square_size+1):     # Horizontal lines
        y = (row-1)*square_size
        canv_pan.create_line(0,y, width, y, fill="grey")
        canv_pan.create_point(0,y)
        canv_pan.create_text(0,y+3,text=f"{y}")
    for col in range(1,width//square_size+1):     # Vertical lines
        x = (col-1)*square_size
        canv_pan.create_line(x,0, x, height, fill="purple")
        canv_pan.create_point(x,0)
        canv_pan.create_text(x+3,3,text=f"{x}")
    # Dots at cross points
    for row in range(1,height//square_size+1):     # Horizontal lines
        y = (row-1)*square_size
        for col in range(1,width//square_size+1):     # Vertical lines
            x = (col-1)*square_size
            canv_pan.create_point(x, y, fill="white")
            if x == y:
                canv_pan.create_text(x+3,y+3,text=f"{x}")
    ### TFD add items to pending
    for item in canv_pan.items:
        canv_pan.add_item(item)
        
    canv_pan.Refresh()
    canv_pan.Update()
    canv_pan.Show()



    def test_msg_before(msg="Before delay"):
        SlTrace.lg(msg)
        
    def test_msg_after(msg="After delay"):
        SlTrace.lg(msg)
    if add_after_test:
        SlTrace.lg("Test after() function")
        SlTrace.setFlags("stdouthasts,decpl=2")    
        delay = 1000    # milliseconds
        SlTrace.lg(f"Delay {delay} milliseconds")
        canv_pan.after(0, test_msg_before)
        canv_pan.after(delay, test_msg_after)

    canv_pan.set_key_press_proc(key_press_proc)
    
    ci = 0
    colors = ["red","orange","yellow",
              "green","blue", "indigo",
              "violet", "white","black"]
    def mouse_left_down_proc(x,y):
        """ Process mouse left down event
        Create point to exercise scaling
        :x,y: grid relative location 
        """
        SlTrace.lg(f"mouse:{x},{y}")
        global ci
        ci += 1
        color = colors[ci%len(colors)]
        screen_pt = canv_pan.get_screen_loc(wx_Point(x,y))
        pt_radius = 4
        canv_pan.create_point(screen_pt.x,screen_pt.y,
                              radius=pt_radius, fill=color)
        canv_pan.create_text(screen_pt.x+1,screen_pt.y+2, text=f"{x},{y}")
        pt_size = pt_radius*2 + 5
        x1,y1 = screen_pt.x,screen_pt.y
        x2,y2 = int(screen_pt.x+pt_size/2),int(screen_pt.y+pt_size/2)
        canv_pan.refresh_rectangle(x1,y1,x2,y2)
        SlTrace.lg(f" mouse:{x},{y} create_point({screen_pt.x},{screen_pt.y})")
        SlTrace.lg(f"panel.sfx,sfy: {canv_pan.sfx},{canv_pan.sfy})")
    canv_pan.set_mouse_left_down_proc(mouse_left_down_proc)
            
    app.MainLoop()
    


# wx_braille_display.py      25Oct2023  crs, from braille_display.py
#                            26Feb2023  crs, add self.x_min, self.y_min
#                            21Feb2023  crs, From braille_display.py
#                            19Apr2022  crs  Author
"""
Display graphics on window    Uses tk.Canvas scanning
Supports simple graphical point, line specification
Supports display of 6-point cells
on grid_width by grid_height grid
Supports writing out braille stream
Uses CanvasGrid for canvas scanning
"""
import os
import sys
import re
import wx
import time
import datetime

import add_src_dir      # Insure site directory path is in sys.path

from select_trace import SlTrace
from wx_speaker_control import SpeakerControlLocal
from braille_cell import BrailleCell
from tk_canvas_grid import TkCanvasGrid
from magnify_info import MagnifyInfo
from wx_audio_draw_window import AudioDrawWindow
from braille_error import BrailleError
from wx_tk_rpc_user import TkRPCUser
from wx_braille_cell_list import BrailleCellList
        
class BrailleDisplay:
    """ Create and display graphics using Braille
    """
    

    def __init__(self,
                 tkr=None,
                 id_title=None,
                 title="Braille Display",
                 src_file="unknown",
                 display_list=None,
                 win_width=800, win_height=800,
                 grid_width=40, grid_height=25,
                 use_full_cells= True,
                 x_min=None, y_min=None,
                 line_width=1, color="black",
                 color_bg = None,
                 color_fill = None,
                 point_resolution=None,
                 blank_char=",",
                 shift_to_edge=None,
                 braille_print=True,
                 braille_window=True,
                 print_cells=False,
                 tk_items=False,
                 canvas_items=False,
                 silent=False,
                 speaker_control=None
                 ):
        """ Setup display
        :wxr
        :tkr: Access to tkinter from remote user
        :id_title: identification title, especially for braille
        :title: display screen title
        :src_file: source file name/path
                if in "src_file() stmt
        :display_list: list of (ix,iy,color) to display
            default: get display from tk
        :win_width: display window width in pixels
            default: 800
        :win_height: display window height in pixels
            default: 800
        :grid_width: braille width in cells
            default: 40
        :grid_height: braille width in cells
            default: 25
        :color: drawing color
                default: turtle default
        :color_fill: fill color
                default: drawing color
        :color_bg: background color
                default: turtle default
        :use_full_cells: Use full cells for point/lines
            e.g. place color letter in cell
            default: True - usefull cells
        :x_min: x value for left side default: -self.draw_width()/2
        :y_min:  y value for bottom default: -self.draw_height()/2
        :line_width: line width
        :point_resolution: Distance between points below
            with, no difference is recognized
            default: computed so as to avoid gaps
                    between connected points
                    conservative to simplify/speed
                    computation
        :blank_char: replacement for non-trailing blanks
                    default "," to provide a "mostly blank",
                    non-compressed blank character for the
                    braille graphics
                    default: "," dot 2.
        :shift_to_edge: shift picture to edge/top
                    to aid in finding
                    default: False - no shift
        :braille_print:  Print text picture targeted for embosser
                    default: True
        :braille_window: Create interactive braille_window
                    default: True
        :print_cells: Print text representation of display cells
                    default: False,
        :tk_items: Print list of tkinter canvas items default: False
        :canvas_items: print whole canvas item info default: False
        :silent: starting val default: False
        :speaker_control: unified sound speach and text control
                default: create
        """
        self.id_title = id_title
        self.title = title
        self.src_file = src_file
        self.display_depth = 0
        if speaker_control is None:
            speaker_control = SpeakerControlLocal()   # access to sound/speech engine
        self.speaker_control = speaker_control
        if tkr is None:
            SlTrace.lg("No link to remote")
            tkr =  TkRPCUser()
        self.tkr = tkr
        if display_list is None:
            cells = tkr.get_cell_specs()
            display_list = BrailleCellList(cells).to_string()
        self.display_list = display_list
        if title is None:
            title = "Braille Display"
        self.title = title
        if win_width is None:
            win_width = 800
        self.win_width = win_width
        if win_height is None:
            win_height = 800
        self.win_height = win_height
        if x_min is None:
            x_min = -self.draw_width()/2
        self.x_min = x_min
        self.x_max = x_min + self.draw_width()
        if y_min is None:
            y_min = -self.draw_height()/2
        self.y_min = y_min
        self.y_max = y_min + self.draw_height()
        
        self.grid_width = grid_width
        self.cell_width = self.draw_width()/self.grid_width
        self.grid_height = grid_height
        self.cell_height = self.draw_height()/self.grid_height
        self._braille_window = braille_window
        self._braille_print = braille_print
        self._print_cells = print_cells
        self.blank_char = blank_char
        if shift_to_edge is None:
            shift_to_edge = False               # TFD
        self.shift_to_edge = shift_to_edge
        self.tk_items = tk_items
        self.canvas_items = canvas_items
        self.app = wx.App()
        SlTrace.lg(f"\nBrailleDisplay: title: {title}\n display_list:{display_list}",
                   "cell_spec,braille_display")
        """
        duplicates from AudioDrawWindow 
        """
    def draw_height(self):
        """ drawing area height reduced from window height
        """
        return self.win_height*(21/25)  # Hack for now

    def draw_width(self):
        """ drawing area width reduced from window width
        """
        return self.win_width*(38/40)  # Hack for now
    
    
    def MainLoop(self):
        """ Just an inclosed access to wx loop
        """
        self.app.MainLoop()
     
    def color_str(self, color):
        """ convert turtle colors arg(s) to color string
        :color: turtle color arg
        """
        return BrailleCell.color_str(color)

    def display(self,
                display_list=None,
                braille_window=True, braille_print=True,
                braille_title=None,
                print_cells=False, title=None,
                points_window=False,
                tk_items=False,
                canvas_items=False,
                silent=False):
        """ display grid
        :display_list: list of (ix,iy,color) cells to display
                default: self.display_list else none
        :braille_window: True - make window display of braille
                        default:True
        :braille_print: True - print braille
                        default: True
        :braille_title: Printed title in braille text output
                        default: Use default title
        :print_cells: True - print out non-empty cells
                        default: False
        :title: text title to display
                    default:None - no title
        :points_window: make window showing display points
                        instead of braille dots
                    default: False - display dots
        :tk_items: True - display tkinter obj in cell
                    default: self.tk_items - False
        :canvas_items: print whole canvas item info
                    default: self.canvas_items - False
        :silent: quiet mode default: False
        """
        SlTrace.lg("self.display()", "braille_display")
        SlTrace.lg(f"""
                   BrailleDisplay.display: title: {title}
                   display_list:{display_list}
                   """, "braille_display")
        self.display_depth += 1
        if self.display_depth > 1:
            self.display_depth -= 1
            return
        if display_list is None:
            display_list = self.display_list
        self.display_list = display_list
        self.braille_title = braille_title
        if hasattr(sys.modules, '__main__'):        
            pgm_file = sys.modules['__main__'].__file__
            file_base_name = os.path.basename(pgm_file)
            current_time = str(datetime.datetime.now())
            mt = re.match(r'(.*):\d+\.\d+$', current_time)
            if mt is not None:
                current_time = mt.group(1)  # Ignore seconds
            username = os.getlogin()
            self.braille_title = f"File: {file_base_name}"
            self.braille_title += f"  Date: {current_time}"
            if username is not None and username != "":
                self.braille_title += f"  User: {username}"
            SlTrace.lg(f"braille_title: {self.braille_title}", "braille_display")
        if self.braille_title is not None:
            title = self.braille_title
        if title is None:
            title = "Audio Feedback -"
        tib = title
        if tib is not None and tib.endswith("-"):
            tib += " Braille Window"
        SlTrace.lg("display - create AudioDrawWindow", "braille_display")
        self.adw = AudioDrawWindow(
                            self.tkr,
                            display_list=self.display_list,
                            app=self.app,
                            id_title=self.id_title,
                            title=title,
                            speaker_control=self.speaker_control,
                            iy0_is_top=True,
                            pgmExit=self.exit,
                            ###x_min=self.x_min, y_min=self.y_min,
                            ###x_max=self.x_max, y_max=self.y_max,
                            silent=silent)
        self.tkr.set_adw(self.adw)  # Link adw to tkr
        
        if braille_print:
            if self.braille_title is not None:
                tib = self.braille_title
            else:
                tib = title
            if tib is not None and tib.endswith("-"):
                tib += " Braille Print Output"
            self.adw.print_braille(title=tib)
        if print_cells:
            tib = title
            if tib is None:
                tib = "Print Cells"
            if tib is not None and tib.endswith("-"):
                tib += " Braille Cells"
            self.adw.print_cells(title=tib)
        if tk_items:
            tib = "tk_items - " + title
            if tib is not None and tib.endswith("-"):
                tib += " Tk Cells"
            self.canvas_grid.show_canvas_items(title=tib)
        if canvas_items:
            tib = "canvas_items - " + title
            if tib is not None and tib.endswith("-"):
                tib += " Canvas items"
            self.canvas_grid.show_canvas(title=tib)
        SlTrace.lg("End of display", "braille_display")
        self.display_depth -= 1
        
        
        
    def snapshot(self, title=None, clear_after=False):
        """ Take snapshot of current braille_screen
        :title: title of snapshot
        :clear_after: clear braille screen after snapshot
        """

    def exit(self, rc=0):
        """ Main program  exit
        """
        SlTrace.lg("BrailleDisplay.exit")
        SlTrace.onexit()    # Force logging quit
        os._exit(rc)

        
        
if __name__ == "__main__":
    import argparse
    from braille_cell_text import BrailleCellText
    from wx_braille_cell_list import BrailleCellList
    from wx_tk_rpc_user import TkRPCUser
    
    spokes_picture="""
    ,,,,,,,,,,,iii
    ,,,,,,,,,,iiiii
    ,,,,,,,,,,iiiii,,,,,,vvv
    ,,,,,,,,,,iiiii,,,,,vvvvv
    ,,,,,,,,,,,,ii,,,,,,vvvvv
    ,,,bb,,,,,,,,i,,,,,,vvvvv
    ,,bbbbb,,,,,,i,,,,,vv
    ,,bbbbb,,,,,,i,,,,vv
    ,,bbbbbbb,,,,ii,,vv
    ,,,,,,,,bbbb,,i,vv,,,,,,,,rr
    ,,,,,,,,,,bbbbivv,,,,,,,,rrrr
    ,,,,,,,,,,,,,bvvrrrrrrrrrrrrr
    ,,,,,,,,,,ggggyoo,,,,,,,,rrrr
    ,,,,,,,,gggg,,y,oo,,,,,,,,rr
    ,,ggggggg,,,,yy,,oo
    ,,ggggg,,,,,,y,,,,oo
    ,,ggggg,,,,,,y,,,,,oo
    ,,,gg,,,,,,,,y,,,,,,ooooo
    ,,,,,,,,,,,,yy,,,,,,ooooo
    ,,,,,,,,,,yyyyy,,,,,ooooo
    ,,,,,,,,,,yyyyy,,,,,,ooo
    ,,,,,,,,,,yyyyy
    ,,,,,,,,,,,yyy
    """
    spokes_bct = BrailleCellText(text=spokes_picture)
    spokes_cells = spokes_bct.get_cells()
    spokes_bcs = BrailleCellList(spokes_cells).to_string()
    
    parser = argparse.ArgumentParser()
    args = parser.parse_args()             # or die "Illegal options"
    SlTrace.lg(f"args: {args}\n")
    tkr = TkRPCUser(simulated=True)

    bd = BrailleDisplay(tkr=tkr, src_file=__file__,
                        display_list=spokes_bcs)
    bd.display(title="Selftest")
    bd.MainLoop()




# wx_trace_control_window.py 19Mar2024  crs, from trace_control_window.py
# Use wxPython
import sys
import time
import atexit

import wx

from select_trace import SlTrace
from wx_trace_control_pan3 import TraceControlPanel
        
class TraceControlWindow(wx.Frame):
    def __init__(self, tcbase=None, change_call=None):
        """ Trace flag dictionary
        :tcbase: - parent - call basis must have tc_destroy to be called if we close
        :change_call: - call with change flag, value, if present
        """
        if tcbase is None:
            SlTrace.lg("Standalone TraceControlWindow")
            root = wx.Frame(None, -1, "Trace Control")
            frame = wx.Frame(root)
            frame.Show()
            self.standalone = True
        self.tcbase = tcbase
        tc_x0 = 800
        tc_y0 = 100
        tc_w = 400
        tc_h = 200
        wx.Frame.__init__(self, tcbase, id=wx.ID_ANY,
                         size=(tc_w,tc_h), pos=(tc_x0,tc_y0))
        self.tcp = TraceControlPanel(self)                    

        if SlTrace.trace("trace_flags"):
            self.list_ckbuttons()
        self.Show()

    def on_exit(self):
        """ Close down window on program exit
        """
        SlTrace.lg("Closing down Trace Control Window")
        self.delete_tc_window()
        
    def update_loop(self):
        """ continue repeated tk.update() calls
        to enable window operation
        """
        loop_time = 50          # Loop recall time (msec)
        self.update()
        self.tc_mw.after(loop_time, self.update_loop)

    def update(self):
        if self.tc_mw is not None and self.tc_mw.winfo_exists():
            self.tc_mw.update()
        
    def sleep(self, sec):
        """ "sleep" for a number of sec
        without stoping tkinter stuff
        :sec: number of milliseconds to delay before returning
        """
        return  # TFD
        if self.tc_mw is None:
            return
        
        self.update()             # Insure at least one update
        now = time.time()
        end_time = now + sec
        while time.time() < end_time:
            if self.tc_mw is None:
                return
            
            self.update()
        return
    
    def mainloop(self):
        self.tc_mw.mainloop()
        
    def delete_tc_window(self):
        """ Process Trace Control window close
        """
        if self.tc_mw is not None:
            self.tc_mw.destroy()
            self.tc_mw = None
        
        if self.tcbase is not None and hasattr(self.tcbase, 'tc_destroy'):
            self.tcbase.tc_destroy()
                
    def enter_entry(self, event):
        flag = self.flag_by_cb[event.widget]
        old_level = SlTrace.getLevel(flag)
        tcb, flag = self.data_by_flag[flag]
                
    def select_button(self, event):
        flag = self.flag_by_cb[event.widget]
        tcb, flag = self.data_by_flag[flag]
        val = SlTrace.getLevel(flag)        # Variable doesn't seem to work for us
        val = not val                           # Keep value in strace
        self.set_trace_level(flag, val, change_cb=False)  # CB already set
        


    def list_ckbuttons(self):
        self.tcp.list_ckbuttons()

if __name__ == '__main__':
    app = wx.App()
    def report_change(flag, val, cklist=None):
        SlTrace.lg("changed: %s = %d" % (flag, val))
        new_val = SlTrace.getLevel(flag)
        SlTrace.lg("New val: %s = %d" % (flag, new_val))
        if cklist is not None:
            cklist.list_ckbuttons()
    
    root = wx.Frame(None)
    SlTrace.set_mw(root)
    ###frame = Frame(root)
    ###frame.pack()
    SlTrace.setProps()
    SlTrace.setFlags("flag1=t,flag2=f,flag3=t,flag4=10, flag5=20. flag6=abc")
    threshold = 5
    SlTrace.setLevel("tint1", 5)
    end_level = 100
    quit_set = False
    def our_quit(flag=None):
        """ Test for traceButton
        :flag: flag arg
        """
        global quit_set
        if flag is None:
            flag = "TBD"
        SlTrace.lg(f"our_quit({flag})")
        SlTrace.report("Quitting Program")
        root.destroy
        quit_set = True
        sys.exit()

        
    def test_int(flag, level=threshold, default=None):
        return  # TFD
        SlTrace.lg(f"Set {flag} to over {end_level} to quit")
        if SlTrace.trace(flag, threshold):
            SlTrace.lg(f"{flag} = {SlTrace.trace(flag)} >= {level}")
        else:
            SlTrace.lg(f"{flag} = {SlTrace.trace(flag)} < {level}")
        if quit_set or SlTrace.trace("tint1") > end_level:
            SlTrace.lg("Manual quit")
            sys.exit()
        SlTrace.traceButton("quit", our_quit)
                    
    tcw = TraceControlWindow(root, change_call=report_change)
    
    test_interval = False
    if test_interval:
        for i in range(end_level):
            test_int("tint1", default=2)
            tcw.sleep(1)
        
    app.MainLoop()    
    SlTrace.lg("End of test")